# Resource object code (Python 3)
# Created by: object code
# Created by: The Resource Compiler for Qt version 6.9.1
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore

qt_resource_data = b"\
\x00\x00[\xec\
@\
font-face{font-f\
amily:KaTeX_AMS;\
font-style:norma\
l;font-weight:40\
0;src:url(qrc://\
/fonts/KaTeX_AMS\
-Regular.woff2) \
format(\x22woff2\x22),\
url(qrc:///fonts\
/KaTeX_AMS-Regul\
ar.woff) format(\
\x22woff\x22),url(qrc:\
///fonts/KaTeX_A\
MS-Regular.ttf) \
format(\x22truetype\
\x22)}@font-face{fo\
nt-family:KaTeX_\
Caligraphic;font\
-style:normal;fo\
nt-weight:700;sr\
c:url(qrc:///fon\
ts/KaTeX_Caligra\
phic-Bold.woff2)\
 format(\x22woff2\x22)\
,url(qrc:///font\
s/KaTeX_Caligrap\
hic-Bold.woff) f\
ormat(\x22woff\x22),ur\
l(qrc:///fonts/K\
aTeX_Caligraphic\
-Bold.ttf) forma\
t(\x22truetype\x22)}@f\
ont-face{font-fa\
mily:KaTeX_Calig\
raphic;font-styl\
e:normal;font-we\
ight:400;src:url\
(qrc:///fonts/Ka\
TeX_Caligraphic-\
Regular.woff2) f\
ormat(\x22woff2\x22),u\
rl(qrc:///fonts/\
KaTeX_Caligraphi\
c-Regular.woff) \
format(\x22woff\x22),u\
rl(qrc:///fonts/\
KaTeX_Caligraphi\
c-Regular.ttf) f\
ormat(\x22truetype\x22\
)}@font-face{fon\
t-family:KaTeX_F\
raktur;font-styl\
e:normal;font-we\
ight:700;src:url\
(qrc:///fonts/Ka\
TeX_Fraktur-Bold\
.woff2) format(\x22\
woff2\x22),url(qrc:\
///fonts/KaTeX_F\
raktur-Bold.woff\
) format(\x22woff\x22)\
,url(qrc:///font\
s/KaTeX_Fraktur-\
Bold.ttf) format\
(\x22truetype\x22)}@fo\
nt-face{font-fam\
ily:KaTeX_Fraktu\
r;font-style:nor\
mal;font-weight:\
400;src:url(qrc:\
///fonts/KaTeX_F\
raktur-Regular.w\
off2) format(\x22wo\
ff2\x22),url(qrc://\
/fonts/KaTeX_Fra\
ktur-Regular.wof\
f) format(\x22woff\x22\
),url(qrc:///fon\
ts/KaTeX_Fraktur\
-Regular.ttf) fo\
rmat(\x22truetype\x22)\
}@font-face{font\
-family:KaTeX_Ma\
in;font-style:no\
rmal;font-weight\
:700;src:url(qrc\
:///fonts/KaTeX_\
Main-Bold.woff2)\
 format(\x22woff2\x22)\
,url(qrc:///font\
s/KaTeX_Main-Bol\
d.woff) format(\x22\
woff\x22),url(qrc:/\
//fonts/KaTeX_Ma\
in-Bold.ttf) for\
mat(\x22truetype\x22)}\
@font-face{font-\
family:KaTeX_Mai\
n;font-style:ita\
lic;font-weight:\
700;src:url(qrc:\
///fonts/KaTeX_M\
ain-BoldItalic.w\
off2) format(\x22wo\
ff2\x22),url(qrc://\
/fonts/KaTeX_Mai\
n-BoldItalic.wof\
f) format(\x22woff\x22\
),url(qrc:///fon\
ts/KaTeX_Main-Bo\
ldItalic.ttf) fo\
rmat(\x22truetype\x22)\
}@font-face{font\
-family:KaTeX_Ma\
in;font-style:it\
alic;font-weight\
:400;src:url(qrc\
:///fonts/KaTeX_\
Main-Italic.woff\
2) format(\x22woff2\
\x22),url(qrc:///fo\
nts/KaTeX_Main-I\
talic.woff) form\
at(\x22woff\x22),url(q\
rc:///fonts/KaTe\
X_Main-Italic.tt\
f) format(\x22truet\
ype\x22)}@font-face\
{font-family:KaT\
eX_Main;font-sty\
le:normal;font-w\
eight:400;src:ur\
l(qrc:///fonts/K\
aTeX_Main-Regula\
r.woff2) format(\
\x22woff2\x22),url(qrc\
:///fonts/KaTeX_\
Main-Regular.wof\
f) format(\x22woff\x22\
),url(qrc:///fon\
ts/KaTeX_Main-Re\
gular.ttf) forma\
t(\x22truetype\x22)}@f\
ont-face{font-fa\
mily:KaTeX_Math;\
font-style:itali\
c;font-weight:70\
0;src:url(qrc://\
/fonts/KaTeX_Mat\
h-BoldItalic.wof\
f2) format(\x22woff\
2\x22),url(qrc:///f\
onts/KaTeX_Math-\
BoldItalic.woff)\
 format(\x22woff\x22),\
url(qrc:///fonts\
/KaTeX_Math-Bold\
Italic.ttf) form\
at(\x22truetype\x22)}@\
font-face{font-f\
amily:KaTeX_Math\
;font-style:ital\
ic;font-weight:4\
00;src:url(qrc:/\
//fonts/KaTeX_Ma\
th-Italic.woff2)\
 format(\x22woff2\x22)\
,url(qrc:///font\
s/KaTeX_Math-Ita\
lic.woff) format\
(\x22woff\x22),url(qrc\
:///fonts/KaTeX_\
Math-Italic.ttf)\
 format(\x22truetyp\
e\x22)}@font-face{f\
ont-family:\x22KaTe\
X_SansSerif\x22;fon\
t-style:normal;f\
ont-weight:700;s\
rc:url(qrc:///fo\
nts/KaTeX_SansSe\
rif-Bold.woff2) \
format(\x22woff2\x22),\
url(qrc:///fonts\
/KaTeX_SansSerif\
-Bold.woff) form\
at(\x22woff\x22),url(q\
rc:///fonts/KaTe\
X_SansSerif-Bold\
.ttf) format(\x22tr\
uetype\x22)}@font-f\
ace{font-family:\
\x22KaTeX_SansSerif\
\x22;font-style:ita\
lic;font-weight:\
400;src:url(qrc:\
///fonts/KaTeX_S\
ansSerif-Italic.\
woff2) format(\x22w\
off2\x22),url(qrc:/\
//fonts/KaTeX_Sa\
nsSerif-Italic.w\
off) format(\x22wof\
f\x22),url(qrc:///f\
onts/KaTeX_SansS\
erif-Italic.ttf)\
 format(\x22truetyp\
e\x22)}@font-face{f\
ont-family:\x22KaTe\
X_SansSerif\x22;fon\
t-style:normal;f\
ont-weight:400;s\
rc:url(qrc:///fo\
nts/KaTeX_SansSe\
rif-Regular.woff\
2) format(\x22woff2\
\x22),url(qrc:///fo\
nts/KaTeX_SansSe\
rif-Regular.woff\
) format(\x22woff\x22)\
,url(qrc:///font\
s/KaTeX_SansSeri\
f-Regular.ttf) f\
ormat(\x22truetype\x22\
)}@font-face{fon\
t-family:KaTeX_S\
cript;font-style\
:normal;font-wei\
ght:400;src:url(\
qrc:///fonts/KaT\
eX_Script-Regula\
r.woff2) format(\
\x22woff2\x22),url(qrc\
:///fonts/KaTeX_\
Script-Regular.w\
off) format(\x22wof\
f\x22),url(qrc:///f\
onts/KaTeX_Scrip\
t-Regular.ttf) f\
ormat(\x22truetype\x22\
)}@font-face{fon\
t-family:KaTeX_S\
ize1;font-style:\
normal;font-weig\
ht:400;src:url(q\
rc:///fonts/KaTe\
X_Size1-Regular.\
woff2) format(\x22w\
off2\x22),url(qrc:/\
//fonts/KaTeX_Si\
ze1-Regular.woff\
) format(\x22woff\x22)\
,url(qrc:///font\
s/KaTeX_Size1-Re\
gular.ttf) forma\
t(\x22truetype\x22)}@f\
ont-face{font-fa\
mily:KaTeX_Size2\
;font-style:norm\
al;font-weight:4\
00;src:url(qrc:/\
//fonts/KaTeX_Si\
ze2-Regular.woff\
2) format(\x22woff2\
\x22),url(qrc:///fo\
nts/KaTeX_Size2-\
Regular.woff) fo\
rmat(\x22woff\x22),url\
(qrc:///fonts/Ka\
TeX_Size2-Regula\
r.ttf) format(\x22t\
ruetype\x22)}@font-\
face{font-family\
:KaTeX_Size3;fon\
t-style:normal;f\
ont-weight:400;s\
rc:url(qrc:///fo\
nts/KaTeX_Size3-\
Regular.woff2) f\
ormat(\x22woff2\x22),u\
rl(qrc:///fonts/\
KaTeX_Size3-Regu\
lar.woff) format\
(\x22woff\x22),url(qrc\
:///fonts/KaTeX_\
Size3-Regular.tt\
f) format(\x22truet\
ype\x22)}@font-face\
{font-family:KaT\
eX_Size4;font-st\
yle:normal;font-\
weight:400;src:u\
rl(qrc:///fonts/\
KaTeX_Size4-Regu\
lar.woff2) forma\
t(\x22woff2\x22),url(q\
rc:///fonts/KaTe\
X_Size4-Regular.\
woff) format(\x22wo\
ff\x22),url(qrc:///\
fonts/KaTeX_Size\
4-Regular.ttf) f\
ormat(\x22truetype\x22\
)}@font-face{fon\
t-family:KaTeX_T\
ypewriter;font-s\
tyle:normal;font\
-weight:400;src:\
url(qrc:///fonts\
/KaTeX_Typewrite\
r-Regular.woff2)\
 format(\x22woff2\x22)\
,url(qrc:///font\
s/KaTeX_Typewrit\
er-Regular.woff)\
 format(\x22woff\x22),\
url(qrc:///fonts\
/KaTeX_Typewrite\
r-Regular.ttf) f\
ormat(\x22truetype\x22\
)}.katex{text-re\
ndering:auto;fon\
t:normal 1.21em \
KaTeX_Main,Times\
 New Roman,serif\
;line-height:1.2\
;text-indent:0}.\
katex *{-ms-high\
-contrast-adjust\
:none!important;\
border-color:cur\
rentColor}.katex\
 .katex-version:\
after{content:\x220\
.16.7\x22}.katex .k\
atex-mathml{clip\
:rect(1px,1px,1p\
x,1px);border:0;\
height:1px;overf\
low:hidden;paddi\
ng:0;position:ab\
solute;width:1px\
}.katex .katex-h\
tml>.newline{dis\
play:block}.kate\
x .base{position\
:relative;white-\
space:nowrap;wid\
th:-webkit-min-c\
ontent;width:-mo\
z-min-content;wi\
dth:min-content}\
.katex .base,.ka\
tex .strut{displ\
ay:inline-block}\
.katex .textbf{f\
ont-weight:700}.\
katex .textit{fo\
nt-style:italic}\
.katex .textrm{f\
ont-family:KaTeX\
_Main}.katex .te\
xtsf{font-family\
:KaTeX_SansSerif\
}.katex .texttt{\
font-family:KaTe\
X_Typewriter}.ka\
tex .mathnormal{\
font-family:KaTe\
X_Math;font-styl\
e:italic}.katex \
.mathit{font-fam\
ily:KaTeX_Main;f\
ont-style:italic\
}.katex .mathrm{\
font-style:norma\
l}.katex .mathbf\
{font-family:KaT\
eX_Main;font-wei\
ght:700}.katex .\
boldsymbol{font-\
family:KaTeX_Mat\
h;font-style:ita\
lic;font-weight:\
700}.katex .amsr\
m,.katex .mathbb\
,.katex .textbb{\
font-family:KaTe\
X_AMS}.katex .ma\
thcal{font-famil\
y:KaTeX_Caligrap\
hic}.katex .math\
frak,.katex .tex\
tfrak{font-famil\
y:KaTeX_Fraktur}\
.katex .mathtt{f\
ont-family:KaTeX\
_Typewriter}.kat\
ex .mathscr,.kat\
ex .textscr{font\
-family:KaTeX_Sc\
ript}.katex .mat\
hsf,.katex .text\
sf{font-family:K\
aTeX_SansSerif}.\
katex .mathbolds\
f,.katex .textbo\
ldsf{font-family\
:KaTeX_SansSerif\
;font-weight:700\
}.katex .mathits\
f,.katex .textit\
sf{font-family:K\
aTeX_SansSerif;f\
ont-style:italic\
}.katex .mainrm{\
font-family:KaTe\
X_Main;font-styl\
e:normal}.katex \
.vlist-t{border-\
collapse:collaps\
e;display:inline\
-table;table-lay\
out:fixed}.katex\
 .vlist-r{displa\
y:table-row}.kat\
ex .vlist{displa\
y:table-cell;pos\
ition:relative;v\
ertical-align:bo\
ttom}.katex .vli\
st>span{display:\
block;height:0;p\
osition:relative\
}.katex .vlist>s\
pan>span{display\
:inline-block}.k\
atex .vlist>span\
>.pstrut{overflo\
w:hidden;width:0\
}.katex .vlist-t\
2{margin-right:-\
2px}.katex .vlis\
t-s{display:tabl\
e-cell;font-size\
:1px;min-width:2\
px;vertical-alig\
n:bottom;width:2\
px}.katex .vbox{\
align-items:base\
line;display:inl\
ine-flex;flex-di\
rection:column}.\
katex .hbox{widt\
h:100%}.katex .h\
box,.katex .thin\
box{display:inli\
ne-flex;flex-dir\
ection:row}.kate\
x .thinbox{max-w\
idth:0;width:0}.\
katex .msupsub{t\
ext-align:left}.\
katex .mfrac>spa\
n>span{text-alig\
n:center}.katex \
.mfrac .frac-lin\
e{border-bottom-\
style:solid;disp\
lay:inline-block\
;width:100%}.kat\
ex .hdashline,.k\
atex .hline,.kat\
ex .mfrac .frac-\
line,.katex .ove\
rline .overline-\
line,.katex .rul\
e,.katex .underl\
ine .underline-l\
ine{min-height:1\
px}.katex .mspac\
e{display:inline\
-block}.katex .c\
lap,.katex .llap\
,.katex .rlap{po\
sition:relative;\
width:0}.katex .\
clap>.inner,.kat\
ex .llap>.inner,\
.katex .rlap>.in\
ner{position:abs\
olute}.katex .cl\
ap>.fix,.katex .\
llap>.fix,.katex\
 .rlap>.fix{disp\
lay:inline-block\
}.katex .llap>.i\
nner{right:0}.ka\
tex .clap>.inner\
,.katex .rlap>.i\
nner{left:0}.kat\
ex .clap>.inner>\
span{margin-left\
:-50%;margin-rig\
ht:50%}.katex .r\
ule{border:0 sol\
id;display:inlin\
e-block;position\
:relative}.katex\
 .hline,.katex .\
overline .overli\
ne-line,.katex .\
underline .under\
line-line{border\
-bottom-style:so\
lid;display:inli\
ne-block;width:1\
00%}.katex .hdas\
hline{border-bot\
tom-style:dashed\
;display:inline-\
block;width:100%\
}.katex .sqrt>.r\
oot{margin-left:\
.27777778em;marg\
in-right:-.55555\
556em}.katex .fo\
ntsize-ensurer.r\
eset-size1.size1\
,.katex .sizing.\
reset-size1.size\
1{font-size:1em}\
.katex .fontsize\
-ensurer.reset-s\
ize1.size2,.kate\
x .sizing.reset-\
size1.size2{font\
-size:1.2em}.kat\
ex .fontsize-ens\
urer.reset-size1\
.size3,.katex .s\
izing.reset-size\
1.size3{font-siz\
e:1.4em}.katex .\
fontsize-ensurer\
.reset-size1.siz\
e4,.katex .sizin\
g.reset-size1.si\
ze4{font-size:1.\
6em}.katex .font\
size-ensurer.res\
et-size1.size5,.\
katex .sizing.re\
set-size1.size5{\
font-size:1.8em}\
.katex .fontsize\
-ensurer.reset-s\
ize1.size6,.kate\
x .sizing.reset-\
size1.size6{font\
-size:2em}.katex\
 .fontsize-ensur\
er.reset-size1.s\
ize7,.katex .siz\
ing.reset-size1.\
size7{font-size:\
2.4em}.katex .fo\
ntsize-ensurer.r\
eset-size1.size8\
,.katex .sizing.\
reset-size1.size\
8{font-size:2.88\
em}.katex .fonts\
ize-ensurer.rese\
t-size1.size9,.k\
atex .sizing.res\
et-size1.size9{f\
ont-size:3.456em\
}.katex .fontsiz\
e-ensurer.reset-\
size1.size10,.ka\
tex .sizing.rese\
t-size1.size10{f\
ont-size:4.148em\
}.katex .fontsiz\
e-ensurer.reset-\
size1.size11,.ka\
tex .sizing.rese\
t-size1.size11{f\
ont-size:4.976em\
}.katex .fontsiz\
e-ensurer.reset-\
size2.size1,.kat\
ex .sizing.reset\
-size2.size1{fon\
t-size:.83333333\
em}.katex .fonts\
ize-ensurer.rese\
t-size2.size2,.k\
atex .sizing.res\
et-size2.size2{f\
ont-size:1em}.ka\
tex .fontsize-en\
surer.reset-size\
2.size3,.katex .\
sizing.reset-siz\
e2.size3{font-si\
ze:1.16666667em}\
.katex .fontsize\
-ensurer.reset-s\
ize2.size4,.kate\
x .sizing.reset-\
size2.size4{font\
-size:1.33333333\
em}.katex .fonts\
ize-ensurer.rese\
t-size2.size5,.k\
atex .sizing.res\
et-size2.size5{f\
ont-size:1.5em}.\
katex .fontsize-\
ensurer.reset-si\
ze2.size6,.katex\
 .sizing.reset-s\
ize2.size6{font-\
size:1.66666667e\
m}.katex .fontsi\
ze-ensurer.reset\
-size2.size7,.ka\
tex .sizing.rese\
t-size2.size7{fo\
nt-size:2em}.kat\
ex .fontsize-ens\
urer.reset-size2\
.size8,.katex .s\
izing.reset-size\
2.size8{font-siz\
e:2.4em}.katex .\
fontsize-ensurer\
.reset-size2.siz\
e9,.katex .sizin\
g.reset-size2.si\
ze9{font-size:2.\
88em}.katex .fon\
tsize-ensurer.re\
set-size2.size10\
,.katex .sizing.\
reset-size2.size\
10{font-size:3.4\
5666667em}.katex\
 .fontsize-ensur\
er.reset-size2.s\
ize11,.katex .si\
zing.reset-size2\
.size11{font-siz\
e:4.14666667em}.\
katex .fontsize-\
ensurer.reset-si\
ze3.size1,.katex\
 .sizing.reset-s\
ize3.size1{font-\
size:.71428571em\
}.katex .fontsiz\
e-ensurer.reset-\
size3.size2,.kat\
ex .sizing.reset\
-size3.size2{fon\
t-size:.85714286\
em}.katex .fonts\
ize-ensurer.rese\
t-size3.size3,.k\
atex .sizing.res\
et-size3.size3{f\
ont-size:1em}.ka\
tex .fontsize-en\
surer.reset-size\
3.size4,.katex .\
sizing.reset-siz\
e3.size4{font-si\
ze:1.14285714em}\
.katex .fontsize\
-ensurer.reset-s\
ize3.size5,.kate\
x .sizing.reset-\
size3.size5{font\
-size:1.28571429\
em}.katex .fonts\
ize-ensurer.rese\
t-size3.size6,.k\
atex .sizing.res\
et-size3.size6{f\
ont-size:1.42857\
143em}.katex .fo\
ntsize-ensurer.r\
eset-size3.size7\
,.katex .sizing.\
reset-size3.size\
7{font-size:1.71\
428571em}.katex \
.fontsize-ensure\
r.reset-size3.si\
ze8,.katex .sizi\
ng.reset-size3.s\
ize8{font-size:2\
.05714286em}.kat\
ex .fontsize-ens\
urer.reset-size3\
.size9,.katex .s\
izing.reset-size\
3.size9{font-siz\
e:2.46857143em}.\
katex .fontsize-\
ensurer.reset-si\
ze3.size10,.kate\
x .sizing.reset-\
size3.size10{fon\
t-size:2.9628571\
4em}.katex .font\
size-ensurer.res\
et-size3.size11,\
.katex .sizing.r\
eset-size3.size1\
1{font-size:3.55\
428571em}.katex \
.fontsize-ensure\
r.reset-size4.si\
ze1,.katex .sizi\
ng.reset-size4.s\
ize1{font-size:.\
625em}.katex .fo\
ntsize-ensurer.r\
eset-size4.size2\
,.katex .sizing.\
reset-size4.size\
2{font-size:.75e\
m}.katex .fontsi\
ze-ensurer.reset\
-size4.size3,.ka\
tex .sizing.rese\
t-size4.size3{fo\
nt-size:.875em}.\
katex .fontsize-\
ensurer.reset-si\
ze4.size4,.katex\
 .sizing.reset-s\
ize4.size4{font-\
size:1em}.katex \
.fontsize-ensure\
r.reset-size4.si\
ze5,.katex .sizi\
ng.reset-size4.s\
ize5{font-size:1\
.125em}.katex .f\
ontsize-ensurer.\
reset-size4.size\
6,.katex .sizing\
.reset-size4.siz\
e6{font-size:1.2\
5em}.katex .font\
size-ensurer.res\
et-size4.size7,.\
katex .sizing.re\
set-size4.size7{\
font-size:1.5em}\
.katex .fontsize\
-ensurer.reset-s\
ize4.size8,.kate\
x .sizing.reset-\
size4.size8{font\
-size:1.8em}.kat\
ex .fontsize-ens\
urer.reset-size4\
.size9,.katex .s\
izing.reset-size\
4.size9{font-siz\
e:2.16em}.katex \
.fontsize-ensure\
r.reset-size4.si\
ze10,.katex .siz\
ing.reset-size4.\
size10{font-size\
:2.5925em}.katex\
 .fontsize-ensur\
er.reset-size4.s\
ize11,.katex .si\
zing.reset-size4\
.size11{font-siz\
e:3.11em}.katex \
.fontsize-ensure\
r.reset-size5.si\
ze1,.katex .sizi\
ng.reset-size5.s\
ize1{font-size:.\
55555556em}.kate\
x .fontsize-ensu\
rer.reset-size5.\
size2,.katex .si\
zing.reset-size5\
.size2{font-size\
:.66666667em}.ka\
tex .fontsize-en\
surer.reset-size\
5.size3,.katex .\
sizing.reset-siz\
e5.size3{font-si\
ze:.77777778em}.\
katex .fontsize-\
ensurer.reset-si\
ze5.size4,.katex\
 .sizing.reset-s\
ize5.size4{font-\
size:.88888889em\
}.katex .fontsiz\
e-ensurer.reset-\
size5.size5,.kat\
ex .sizing.reset\
-size5.size5{fon\
t-size:1em}.kate\
x .fontsize-ensu\
rer.reset-size5.\
size6,.katex .si\
zing.reset-size5\
.size6{font-size\
:1.11111111em}.k\
atex .fontsize-e\
nsurer.reset-siz\
e5.size7,.katex \
.sizing.reset-si\
ze5.size7{font-s\
ize:1.33333333em\
}.katex .fontsiz\
e-ensurer.reset-\
size5.size8,.kat\
ex .sizing.reset\
-size5.size8{fon\
t-size:1.6em}.ka\
tex .fontsize-en\
surer.reset-size\
5.size9,.katex .\
sizing.reset-siz\
e5.size9{font-si\
ze:1.92em}.katex\
 .fontsize-ensur\
er.reset-size5.s\
ize10,.katex .si\
zing.reset-size5\
.size10{font-siz\
e:2.30444444em}.\
katex .fontsize-\
ensurer.reset-si\
ze5.size11,.kate\
x .sizing.reset-\
size5.size11{fon\
t-size:2.7644444\
4em}.katex .font\
size-ensurer.res\
et-size6.size1,.\
katex .sizing.re\
set-size6.size1{\
font-size:.5em}.\
katex .fontsize-\
ensurer.reset-si\
ze6.size2,.katex\
 .sizing.reset-s\
ize6.size2{font-\
size:.6em}.katex\
 .fontsize-ensur\
er.reset-size6.s\
ize3,.katex .siz\
ing.reset-size6.\
size3{font-size:\
.7em}.katex .fon\
tsize-ensurer.re\
set-size6.size4,\
.katex .sizing.r\
eset-size6.size4\
{font-size:.8em}\
.katex .fontsize\
-ensurer.reset-s\
ize6.size5,.kate\
x .sizing.reset-\
size6.size5{font\
-size:.9em}.kate\
x .fontsize-ensu\
rer.reset-size6.\
size6,.katex .si\
zing.reset-size6\
.size6{font-size\
:1em}.katex .fon\
tsize-ensurer.re\
set-size6.size7,\
.katex .sizing.r\
eset-size6.size7\
{font-size:1.2em\
}.katex .fontsiz\
e-ensurer.reset-\
size6.size8,.kat\
ex .sizing.reset\
-size6.size8{fon\
t-size:1.44em}.k\
atex .fontsize-e\
nsurer.reset-siz\
e6.size9,.katex \
.sizing.reset-si\
ze6.size9{font-s\
ize:1.728em}.kat\
ex .fontsize-ens\
urer.reset-size6\
.size10,.katex .\
sizing.reset-siz\
e6.size10{font-s\
ize:2.074em}.kat\
ex .fontsize-ens\
urer.reset-size6\
.size11,.katex .\
sizing.reset-siz\
e6.size11{font-s\
ize:2.488em}.kat\
ex .fontsize-ens\
urer.reset-size7\
.size1,.katex .s\
izing.reset-size\
7.size1{font-siz\
e:.41666667em}.k\
atex .fontsize-e\
nsurer.reset-siz\
e7.size2,.katex \
.sizing.reset-si\
ze7.size2{font-s\
ize:.5em}.katex \
.fontsize-ensure\
r.reset-size7.si\
ze3,.katex .sizi\
ng.reset-size7.s\
ize3{font-size:.\
58333333em}.kate\
x .fontsize-ensu\
rer.reset-size7.\
size4,.katex .si\
zing.reset-size7\
.size4{font-size\
:.66666667em}.ka\
tex .fontsize-en\
surer.reset-size\
7.size5,.katex .\
sizing.reset-siz\
e7.size5{font-si\
ze:.75em}.katex \
.fontsize-ensure\
r.reset-size7.si\
ze6,.katex .sizi\
ng.reset-size7.s\
ize6{font-size:.\
83333333em}.kate\
x .fontsize-ensu\
rer.reset-size7.\
size7,.katex .si\
zing.reset-size7\
.size7{font-size\
:1em}.katex .fon\
tsize-ensurer.re\
set-size7.size8,\
.katex .sizing.r\
eset-size7.size8\
{font-size:1.2em\
}.katex .fontsiz\
e-ensurer.reset-\
size7.size9,.kat\
ex .sizing.reset\
-size7.size9{fon\
t-size:1.44em}.k\
atex .fontsize-e\
nsurer.reset-siz\
e7.size10,.katex\
 .sizing.reset-s\
ize7.size10{font\
-size:1.72833333\
em}.katex .fonts\
ize-ensurer.rese\
t-size7.size11,.\
katex .sizing.re\
set-size7.size11\
{font-size:2.073\
33333em}.katex .\
fontsize-ensurer\
.reset-size8.siz\
e1,.katex .sizin\
g.reset-size8.si\
ze1{font-size:.3\
4722222em}.katex\
 .fontsize-ensur\
er.reset-size8.s\
ize2,.katex .siz\
ing.reset-size8.\
size2{font-size:\
.41666667em}.kat\
ex .fontsize-ens\
urer.reset-size8\
.size3,.katex .s\
izing.reset-size\
8.size3{font-siz\
e:.48611111em}.k\
atex .fontsize-e\
nsurer.reset-siz\
e8.size4,.katex \
.sizing.reset-si\
ze8.size4{font-s\
ize:.55555556em}\
.katex .fontsize\
-ensurer.reset-s\
ize8.size5,.kate\
x .sizing.reset-\
size8.size5{font\
-size:.625em}.ka\
tex .fontsize-en\
surer.reset-size\
8.size6,.katex .\
sizing.reset-siz\
e8.size6{font-si\
ze:.69444444em}.\
katex .fontsize-\
ensurer.reset-si\
ze8.size7,.katex\
 .sizing.reset-s\
ize8.size7{font-\
size:.83333333em\
}.katex .fontsiz\
e-ensurer.reset-\
size8.size8,.kat\
ex .sizing.reset\
-size8.size8{fon\
t-size:1em}.kate\
x .fontsize-ensu\
rer.reset-size8.\
size9,.katex .si\
zing.reset-size8\
.size9{font-size\
:1.2em}.katex .f\
ontsize-ensurer.\
reset-size8.size\
10,.katex .sizin\
g.reset-size8.si\
ze10{font-size:1\
.44027778em}.kat\
ex .fontsize-ens\
urer.reset-size8\
.size11,.katex .\
sizing.reset-siz\
e8.size11{font-s\
ize:1.72777778em\
}.katex .fontsiz\
e-ensurer.reset-\
size9.size1,.kat\
ex .sizing.reset\
-size9.size1{fon\
t-size:.28935185\
em}.katex .fonts\
ize-ensurer.rese\
t-size9.size2,.k\
atex .sizing.res\
et-size9.size2{f\
ont-size:.347222\
22em}.katex .fon\
tsize-ensurer.re\
set-size9.size3,\
.katex .sizing.r\
eset-size9.size3\
{font-size:.4050\
9259em}.katex .f\
ontsize-ensurer.\
reset-size9.size\
4,.katex .sizing\
.reset-size9.siz\
e4{font-size:.46\
296296em}.katex \
.fontsize-ensure\
r.reset-size9.si\
ze5,.katex .sizi\
ng.reset-size9.s\
ize5{font-size:.\
52083333em}.kate\
x .fontsize-ensu\
rer.reset-size9.\
size6,.katex .si\
zing.reset-size9\
.size6{font-size\
:.5787037em}.kat\
ex .fontsize-ens\
urer.reset-size9\
.size7,.katex .s\
izing.reset-size\
9.size7{font-siz\
e:.69444444em}.k\
atex .fontsize-e\
nsurer.reset-siz\
e9.size8,.katex \
.sizing.reset-si\
ze9.size8{font-s\
ize:.83333333em}\
.katex .fontsize\
-ensurer.reset-s\
ize9.size9,.kate\
x .sizing.reset-\
size9.size9{font\
-size:1em}.katex\
 .fontsize-ensur\
er.reset-size9.s\
ize10,.katex .si\
zing.reset-size9\
.size10{font-siz\
e:1.20023148em}.\
katex .fontsize-\
ensurer.reset-si\
ze9.size11,.kate\
x .sizing.reset-\
size9.size11{fon\
t-size:1.4398148\
1em}.katex .font\
size-ensurer.res\
et-size10.size1,\
.katex .sizing.r\
eset-size10.size\
1{font-size:.241\
08004em}.katex .\
fontsize-ensurer\
.reset-size10.si\
ze2,.katex .sizi\
ng.reset-size10.\
size2{font-size:\
.28929605em}.kat\
ex .fontsize-ens\
urer.reset-size1\
0.size3,.katex .\
sizing.reset-siz\
e10.size3{font-s\
ize:.33751205em}\
.katex .fontsize\
-ensurer.reset-s\
ize10.size4,.kat\
ex .sizing.reset\
-size10.size4{fo\
nt-size:.3857280\
6em}.katex .font\
size-ensurer.res\
et-size10.size5,\
.katex .sizing.r\
eset-size10.size\
5{font-size:.433\
94407em}.katex .\
fontsize-ensurer\
.reset-size10.si\
ze6,.katex .sizi\
ng.reset-size10.\
size6{font-size:\
.48216008em}.kat\
ex .fontsize-ens\
urer.reset-size1\
0.size7,.katex .\
sizing.reset-siz\
e10.size7{font-s\
ize:.57859209em}\
.katex .fontsize\
-ensurer.reset-s\
ize10.size8,.kat\
ex .sizing.reset\
-size10.size8{fo\
nt-size:.6943105\
1em}.katex .font\
size-ensurer.res\
et-size10.size9,\
.katex .sizing.r\
eset-size10.size\
9{font-size:.833\
17261em}.katex .\
fontsize-ensurer\
.reset-size10.si\
ze10,.katex .siz\
ing.reset-size10\
.size10{font-siz\
e:1em}.katex .fo\
ntsize-ensurer.r\
eset-size10.size\
11,.katex .sizin\
g.reset-size10.s\
ize11{font-size:\
1.19961427em}.ka\
tex .fontsize-en\
surer.reset-size\
11.size1,.katex \
.sizing.reset-si\
ze11.size1{font-\
size:.20096463em\
}.katex .fontsiz\
e-ensurer.reset-\
size11.size2,.ka\
tex .sizing.rese\
t-size11.size2{f\
ont-size:.241157\
56em}.katex .fon\
tsize-ensurer.re\
set-size11.size3\
,.katex .sizing.\
reset-size11.siz\
e3{font-size:.28\
135048em}.katex \
.fontsize-ensure\
r.reset-size11.s\
ize4,.katex .siz\
ing.reset-size11\
.size4{font-size\
:.32154341em}.ka\
tex .fontsize-en\
surer.reset-size\
11.size5,.katex \
.sizing.reset-si\
ze11.size5{font-\
size:.36173633em\
}.katex .fontsiz\
e-ensurer.reset-\
size11.size6,.ka\
tex .sizing.rese\
t-size11.size6{f\
ont-size:.401929\
26em}.katex .fon\
tsize-ensurer.re\
set-size11.size7\
,.katex .sizing.\
reset-size11.siz\
e7{font-size:.48\
231511em}.katex \
.fontsize-ensure\
r.reset-size11.s\
ize8,.katex .siz\
ing.reset-size11\
.size8{font-size\
:.57877814em}.ka\
tex .fontsize-en\
surer.reset-size\
11.size9,.katex \
.sizing.reset-si\
ze11.size9{font-\
size:.69453376em\
}.katex .fontsiz\
e-ensurer.reset-\
size11.size10,.k\
atex .sizing.res\
et-size11.size10\
{font-size:.8336\
0129em}.katex .f\
ontsize-ensurer.\
reset-size11.siz\
e11,.katex .sizi\
ng.reset-size11.\
size11{font-size\
:1em}.katex .del\
imsizing.size1{f\
ont-family:KaTeX\
_Size1}.katex .d\
elimsizing.size2\
{font-family:KaT\
eX_Size2}.katex \
.delimsizing.siz\
e3{font-family:K\
aTeX_Size3}.kate\
x .delimsizing.s\
ize4{font-family\
:KaTeX_Size4}.ka\
tex .delimsizing\
.mult .delim-siz\
e1>span{font-fam\
ily:KaTeX_Size1}\
.katex .delimsiz\
ing.mult .delim-\
size4>span{font-\
family:KaTeX_Siz\
e4}.katex .nulld\
elimiter{display\
:inline-block;wi\
dth:.12em}.katex\
 .delimcenter,.k\
atex .op-symbol{\
position:relativ\
e}.katex .op-sym\
bol.small-op{fon\
t-family:KaTeX_S\
ize1}.katex .op-\
symbol.large-op{\
font-family:KaTe\
X_Size2}.katex .\
accent>.vlist-t,\
.katex .op-limit\
s>.vlist-t{text-\
align:center}.ka\
tex .accent .acc\
ent-body{positio\
n:relative}.kate\
x .accent .accen\
t-body:not(.acce\
nt-full){width:0\
}.katex .overlay\
{display:block}.\
katex .mtable .v\
ertical-separato\
r{display:inline\
-block;min-width\
:1px}.katex .mta\
ble .arraycolsep\
{display:inline-\
block}.katex .mt\
able .col-align-\
c>.vlist-t{text-\
align:center}.ka\
tex .mtable .col\
-align-l>.vlist-\
t{text-align:lef\
t}.katex .mtable\
 .col-align-r>.v\
list-t{text-alig\
n:right}.katex .\
svg-align{text-a\
lign:left}.katex\
 svg{fill:curren\
tColor;stroke:cu\
rrentColor;fill-\
rule:nonzero;fil\
l-opacity:1;stro\
ke-width:1;strok\
e-linecap:butt;s\
troke-linejoin:m\
iter;stroke-mite\
rlimit:4;stroke-\
dasharray:none;s\
troke-dashoffset\
:0;stroke-opacit\
y:1;display:bloc\
k;height:inherit\
;position:absolu\
te;width:100%}.k\
atex svg path{st\
roke:none}.katex\
 img{border-styl\
e:none;max-heigh\
t:none;max-width\
:none;min-height\
:0;min-width:0}.\
katex .stretchy{\
display:block;ov\
erflow:hidden;po\
sition:relative;\
width:100%}.kate\
x .stretchy:afte\
r,.katex .stretc\
hy:before{conten\
t:\x22\x22}.katex .hid\
e-tail{overflow:\
hidden;position:\
relative;width:1\
00%}.katex .half\
arrow-left{left:\
0;overflow:hidde\
n;position:absol\
ute;width:50.2%}\
.katex .halfarro\
w-right{overflow\
:hidden;position\
:absolute;right:\
0;width:50.2%}.k\
atex .brace-left\
{left:0;overflow\
:hidden;position\
:absolute;width:\
25.1%}.katex .br\
ace-center{left:\
25%;overflow:hid\
den;position:abs\
olute;width:50%}\
.katex .brace-ri\
ght{overflow:hid\
den;position:abs\
olute;right:0;wi\
dth:25.1%}.katex\
 .x-arrow-pad{pa\
dding:0 .5em}.ka\
tex .cd-arrow-pa\
d{padding:0 .555\
56em 0 .27778em}\
.katex .mover,.k\
atex .munder,.ka\
tex .x-arrow{tex\
t-align:center}.\
katex .boxpad{pa\
dding:0 .3em}.ka\
tex .fbox,.katex\
 .fcolorbox{bord\
er:.04em solid;b\
ox-sizing:border\
-box}.katex .can\
cel-pad{padding:\
0 .2em}.katex .c\
ancel-lap{margin\
-left:-.2em;marg\
in-right:-.2em}.\
katex .sout{bord\
er-bottom-style:\
solid;border-bot\
tom-width:.08em}\
.katex .angl{bor\
der-right:.049em\
 solid;border-to\
p:.049em solid;b\
ox-sizing:border\
-box;margin-righ\
t:.03889em}.kate\
x .anglpad{paddi\
ng:0 .03889em}.k\
atex .eqn-num:be\
fore{content:\x22(\x22\
 counter(katexEq\
nNo) \x22)\x22;counter\
-increment:katex\
EqnNo}.katex .mm\
l-eqn-num:before\
{content:\x22(\x22 cou\
nter(mmlEqnNo) \x22\
)\x22;counter-incre\
ment:mmlEqnNo}.k\
atex .mtr-glue{w\
idth:50%}.katex \
.cd-vert-arrow{d\
isplay:inline-bl\
ock;position:rel\
ative}.katex .cd\
-label-left{disp\
lay:inline-block\
;position:absolu\
te;right:calc(50\
% + .3em);text-a\
lign:left}.katex\
 .cd-label-right\
{display:inline-\
block;left:calc(\
50% + .3em);posi\
tion:absolute;te\
xt-align:right}.\
katex-display{di\
splay:block;marg\
in:1em 0;text-al\
ign:center}.kate\
x-display>.katex\
{display:block;t\
ext-align:center\
;white-space:now\
rap}.katex-displ\
ay>.katex>.katex\
-html{display:bl\
ock;position:rel\
ative}.katex-dis\
play>.katex>.kat\
ex-html>.tag{pos\
ition:absolute;r\
ight:0}.katex-di\
splay.leqno>.kat\
ex>.katex-html>.\
tag{left:0;right\
:auto}.katex-dis\
play.fleqn>.kate\
x{padding-left:2\
em;text-align:le\
ft}body{counter-\
reset:katexEqnNo\
 mmlEqnNo}\x0a\
"

qt_resource_name = b"\
\x00\x03\
\x00\x00j\xa3\
\x00c\
\x00s\x00s\
\x00\x0d\
\x01\x14\xdd\xc3\
\x00k\
\x00a\x00t\x00e\x00x\x00.\x00m\x00i\x00n\x00.\x00c\x00s\x00s\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x02\
\x00\x00\x00\x0c\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\
"

def qInitResources():
    QtCore.qRegisterResourceData(0x01, qt_resource_struct, qt_resource_name, qt_resource_data)

def qCleanupResources():
    QtCore.qUnregisterResourceData(0x01, qt_resource_struct, qt_resource_name, qt_resource_data)

qInitResources()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.07 05:00:00                  #
# ================================================== #

import json
import re
import html

class Helpers:

    _RE_HTML_ANGLE_OR_MATH = re.compile(r'(\\\[.*?\\\])|(<)|(>)', flags=re.DOTALL)
    _RE_WORKDIR_TOKEN = re.compile(r'\(%workdir%([^)]+)\)')
    _RE_APPDIR_TOKEN = re.compile(r'\(%appdir%([^)]+)\)')

    _PLACEHOLDER_THINK_OPEN = "{{{{think}}}}"
    _PLACEHOLDER_THINK_CLOSE = "{{{{/think}}}}"

    #_RE_TOOL_TAG = re.compile(r"&lt;tool&gt;(.*?)&lt;/tool&gt;", re.DOTALL)
    _RE_TOOL_TAG = re.compile(r"<tool>(.*?)</tool>", re.DOTALL)
    _RE_THINK_TAG = re.compile(r"<think>(.*?)</think>", re.DOTALL)
    _RE_EXECUTE_TAG = re.compile(r"<execute>(.*?)</execute>", re.DOTALL)
    #_RE_MATH_PARENS = re.compile(r"\\\((.*?)\\\)", re.DOTALL)

    def __init__(self, window=None):
        """
        Helpers

        :param window: Window instance
        """
        self.window = window

    def _html_escape_keep_math(self, m: re.Match) -> str:
        """
        Replaces < and > with &lt; / &gt;, leaving \\[ ... \\]

        :param m: regex match object
        :return: escaped string
        """
        g1 = m.group(1)
        if g1 is not None:
            return g1  # fragment of formula \[ ... \]
        return "&lt;" if m.group(2) is not None else "&gt;"

    def _unescape_lt_gt(self, s: str) -> str:
        """
        Unescape &lt; and &gt; in a string

        :param s: string to unescape
        :return: unescaped string
        """
        if "&lt;" not in s and "&gt;" not in s:
            return s
        if "&lt;" in s:
            s = s.replace("&lt;", "<")
        if "&gt;" in s:
            s = s.replace("&gt;", ">")
        return s

    def _repl_tool_cmd(self, m: re.Match) -> str:
        """
        Replace tool command tags with HTML paragraph

        :param m: regex match object
        :return: formatted HTML string
        """
        return f'[!cmd]{html.escape(m.group(1))}[/!cmd]'

    def _repl_think(self, m: re.Match) -> str:
        """
        Replace think tags with HTML paragraph

        :param m: regex match object
        :return: formatted HTML string
        """
        g = m.group(1).replace("\n", "<br>")
        return f'[!think]{html.escape(g)}[/!think]'

    def _repl_execute(self, m: re.Match) -> str:
        """
        Replace execute tags with HTML paragraph

        :param m: regex match object
        :return: formatted HTML string
        """
        return f'[!exec]{html.escape(m.group(1))}[/!exec]'

    def _repl_math_fix(self, m: re.Match) -> str:
        """
        Fix math formula by replacing &lt; and &gt; with < and > inside \\( ... \\)

        :param m: regex match object
        :return: formatted math string
        """
        inner = m.group(1)
        if "&lt;" not in inner and "&gt;" not in inner:
            return m.group(0)
        inner = inner.replace("&lt;", "<").replace("&gt;", ">")
        return f"\\({inner}\\)"

    def replace_code_tags(self, text: str) -> str:
        """
        Replace cmd code tags

        :param text:
        :return: replaced text
        """
        s = text

        # --- tool tags ---
        if "<tool>" in s and "</tool>" in s:
            s = self._RE_TOOL_TAG.sub(self._repl_tool_cmd, s)

        # --- fix math formula \( ... \) ---
        #if "\\(" in s and "\\)" in s and ("&lt;" in s or "&gt;" in s):
            #s = self._RE_MATH_PARENS.sub(self._repl_math_fix, s)

        return s

    def replace_think_tags(self, text: str) -> str:
        """
        Replace think tags

        :param text:
        :return: replaced text
        """
        s = text

        # --- think tags ---
        if "<think>" in s and "</think>" in s:
            s = self._RE_THINK_TAG.sub(self._repl_think, s)

        return s

    def replace_execute_tags(self, text: str) -> str:
        """
        Replace execute tags

        :param text:
        :return: replaced text
        """
        s = text

        # --- execute tags ---
        if "<execute>" in s and "</execute>" in s:
            s = self._RE_EXECUTE_TAG.sub(self._repl_execute, s)

        return s

    def pre_format_text(self, text: str) -> str:
        """
        Pre-format text

        :param text: text to format
        :return: formatted text
        """
        """
        if "<" in s or ">" in s:
            had_think = ("<think>" in s) or ("</think>" in s)
            if had_think:
                s = s.replace("<think>", self._PLACEHOLDER_THINK_OPEN).replace("</think>", self._PLACEHOLDER_THINK_CLOSE)

            s = self._RE_HTML_ANGLE_OR_MATH.sub(self._html_escape_keep_math, s)

            if had_think:
                s = s.replace(self._PLACEHOLDER_THINK_OPEN, "<think>").replace(self._PLACEHOLDER_THINK_CLOSE, "</think>")
                if "<think>\n" in s:
                    s = s.replace("<think>\n", "<think>")
        """

        # replace tags with markdown placeholders (will be converted to HTML in JS runtime)
        s = self.replace_code_tags(text.strip())
        s = self.replace_think_tags(s)
        s = self.replace_execute_tags(s)

        # replace workdir token
        if "%workdir%" in s:
            prefix = self.window.core.filesystem.get_workdir_prefix()
            s = self._RE_WORKDIR_TOKEN.sub(lambda m, p=prefix: f'({p}{m.group(1)})', s)

        if "%appdir%" in s:
            prefix = self.window.core.config.get_app_path()
            s = self._RE_APPDIR_TOKEN.sub(lambda m, p=prefix: f'({p}{m.group(1)})', s)

        return s

    def post_format_text(self, text: str) -> str:
        """
        Post-format text

        :param text: text to format
        :return: formatted text
        """
        s = text
        if self.window.core.config.get("agent.output.render.all", False):
            if "__agent_begin__" in s or "__agent_end__" in s:
                s = s.replace("__agent_begin__", '<div class="msg-agent">').replace("__agent_end__", "</div>")
        return s.strip()

    def format_user_text(self, text: str) -> str:
        """
        Post-format user text

        :param text: text to format
        :return: formatted text
        """
        s = html.escape(text).replace("\n", "<br>")

        trimmed = s.strip()
        if trimmed.endswith("]") and (trimmed.startswith("[") or trimmed.startswith("&gt; [")):
            s = f'<div class="cmd">&gt; {s}</div>'
        return s

    def format_cmd_text(self, text: str, indent: bool = False) -> str:
        """
        Post-format cmd text

        :param text: text to format
        :param indent: whether to indent text
        :return: formatted text
        """
        if not text:
            return ""
        if not indent:
            return html.escape(text)
        else:
            try:
                return html.escape(json.dumps(json.loads(text), indent=2))
            except Exception:
                return html.escape(text)


    def format_chunk(self, text: str) -> str:
        """
        Format chunk

        :param text: text to format
        :return: formatted text
        """
        return text
        # return text.replace("\n", "<br/>")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.15 01:00:00                  #
# ================================================== #

import os
from typing import Optional, List, Dict

from pygpt_net.core.types import (
    MODE_LLAMA_INDEX,
)
from llama_index.core.llms.llm import BaseLLM as LlamaBaseLLM
from llama_index.core.base.embeddings.base import BaseEmbedding

from pygpt_net.provider.llms.base import BaseLLM
from pygpt_net.item.model import ModelItem


class MistralAILLM(BaseLLM):
    def __init__(self, *args, **kwargs):
        super(MistralAILLM, self).__init__(*args, **kwargs)
        self.id = "mistral_ai"
        self.name = "Mistral AI"
        self.type = [MODE_LLAMA_INDEX, "embeddings"]

    def llama(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ) -> LlamaBaseLLM:
        """
        Return LLM provider instance for llama

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: LLM provider instance
        """
        from llama_index.llms.mistralai import MistralAI
        class MistralAIWithProxy(MistralAI):
            def __init__(self, *args, proxy: Optional[str] = None, **kwargs):
                endpoint = kwargs.get("endpoint")
                super().__init__(*args, **kwargs)
                if not proxy:
                    return

                import httpx
                from mistralai import Mistral
                timeout = getattr(self, "timeout", 120)

                try:
                    sync_client = httpx.Client(proxy=proxy, timeout=timeout, follow_redirects=True)
                    async_client = httpx.AsyncClient(proxy=proxy, timeout=timeout, follow_redirects=True)
                except TypeError:
                    sync_client = httpx.Client(proxies=proxy, timeout=timeout, follow_redirects=True)
                    async_client = httpx.AsyncClient(proxies=proxy, timeout=timeout, follow_redirects=True)

                sdk_kwargs = {
                    "api_key": self.api_key,
                    "client": sync_client,
                    "async_client": async_client,
                }
                if endpoint:
                    sdk_kwargs["server_url"] = endpoint

                self._client = Mistral(**sdk_kwargs)

        args = self.parse_args(model.llama_index, window)
        proxy = window.core.config.get("api_proxy") or None
        if not window.core.config.get("api_proxy.enabled", False):
            proxy = None
        if "model" not in args:
            args["model"] = model.id
        if "api_key" not in args or args["api_key"] == "":
            args["api_key"] = window.core.config.get("api_key_mistral", "")
        return MistralAIWithProxy(**args, proxy=proxy)

    def get_embeddings_model(
            self,
            window,
            config: Optional[List[Dict]] = None
    ) -> BaseEmbedding:
        """
        Return provider instance for embeddings

        :param window: window instance
        :param config: config keyword arguments list
        :return: Embedding provider instance
        """
        from llama_index.embeddings.mistralai import MistralAIEmbedding
        class MistralAIEmbeddingWithProxy(MistralAIEmbedding):
            def __init__(self, *args, proxy: Optional[str] = None, api_key: Optional[str] = None, **kwargs):
                captured_key = api_key or os.environ.get("MISTRAL_API_KEY", "")
                super().__init__(*args, api_key=api_key, **kwargs)

                if not proxy:
                    return

                import httpx
                try:
                    sync_client = httpx.Client(proxy=proxy, timeout=60.0, follow_redirects=True)
                    async_client = httpx.AsyncClient(proxy=proxy, timeout=60.0, follow_redirects=True)
                except TypeError:
                    sync_client = httpx.Client(proxies=proxy, timeout=60.0, follow_redirects=True)
                    async_client = httpx.AsyncClient(proxies=proxy, timeout=60.0, follow_redirects=True)

                from mistralai import Mistral
                server_url = os.environ.get("MISTRAL_ENDPOINT") or None
                self._client = Mistral(
                    api_key=captured_key,
                    client=sync_client,
                    async_client=async_client,
                    **({"server_url": server_url} if server_url else {}),
                )
                if hasattr(self, "_mistralai_client"):
                    self._mistralai_client = self._client
        args = {}
        if config is not None:
            args = self.parse_args({
                "args": config,
            }, window)
        if "api_key" not in args or args["api_key"] == "":
            args["api_key"] = window.core.config.get("api_key_mistral", "")
        if "model" in args and "model_name" not in args:
            args["model_name"] = args.pop("model")

        proxy = window.core.config.get("api_proxy") or None
        if not window.core.config.get("api_proxy.enabled", False):
            proxy = None
        return MistralAIEmbeddingWithProxy(**args, proxy=proxy)

    def init_embeddings(
            self,
            window,
            env: Optional[List[Dict]] = None
    ):
        """
        Initialize embeddings provider

        :param window: window instance
        :param env: ENV configuration list
        """
        super(MistralAILLM, self).init_embeddings(window, env)

        # === FIX FOR LOCAL EMBEDDINGS ===
        # if there is no OpenAI api key then set fake key to prevent empty key Llama-index error
        if ('OPENAI_API_KEY' not in os.environ
                and (window.core.config.get('api_key') is None or window.core.config.get('api_key') == "")):
            os.environ['OPENAI_API_KEY'] = "_"

    def get_models(
            self,
            window,
    ) -> List[Dict]:
        """
        Return list of models for the provider

        :param window: window instance
        :return: list of models
        """
        items = []
        client = self.get_client(window)
        models_list = client.models.list()
        if models_list.data:
            for item in models_list.data:
                id = item.id
                items.append({
                    "id": id,
                    "name": id,
                })
        return items



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.12.27 21:00:00                  #
# ================================================== #

import copy
import os
from typing import List, Dict, Optional, Any

from PySide6 import QtCore
from PySide6.QtWidgets import QApplication

from pygpt_net.core.events import Event
from pygpt_net.core.types import (
    MODE_CHAT,
    MODE_LLAMA_INDEX,
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_AGENT_OPENAI,
    MODE_EXPERT,
)
from pygpt_net.utils import trans


class Importer:

    def __init__(self, window=None):
        """
        Models importer controller

        :param window: Window instance
        """
        self.window = window
        self.dialog = False
        self.initialized = False
        self.width = 800
        self.height = 500
        self.items_available_all = {}  # available models for import
        self.items_available = {}  # available models for import
        self.items_current = {}  # current models in use
        self.pending = {}  # waiting to be imported models
        self.removed = {}  # waiting to be removed models
        self.selected_available = None  # selected available model (single)
        self.selected_current = None  # selected current model (single)
        self.selected_available_ids: List[str] = []  # multi-selected available models
        self.selected_current_ids: List[str] = []  # multi-selected current models
        self.all = False  # show all models, not only available for import
        self.provider = "_"  # default provider

    def in_current(self, model: str) -> bool:
        """
        Check if model is in current list

        :param model: model ID
        :return: True if model is in current list, False otherwise
        """
        if self.provider not in self.items_current:
            self.items_current[self.provider] = {}
        if model in self.items_current[self.provider]:
            return True
        for key in list(self.items_current[self.provider].keys()):
            m = self.items_current[self.provider][key]
            if m.id == model:  # also check in IDs
                return True
        return False

    def change_available(self):
        """On change available model selection"""
        view = self.window.ui.nodes["models.importer.available"]
        sel_model = view.selectionModel()
        rows = sel_model.selectedRows() if sel_model else []
        # collect selected model keys by reading stored tooltip role (stable ID)
        selected_ids = []
        for ix in rows:
            key = ix.data(QtCore.Qt.ToolTipRole)
            if not key:
                key = self.get_by_idx(ix.row(), self.items_available)
            if key and key not in selected_ids:
                selected_ids.append(key)

        self.selected_available_ids = selected_ids
        if not selected_ids:
            self.selected_available = None
            self.window.ui.nodes["models.importer.add"].setEnabled(False)
            return

        # keep a single reference for backward compatibility (last selected)
        self.selected_available = selected_ids[-1]

        # enable add if any of selected can be added (not already in current)
        can_add = False
        for key in selected_ids:
            if key is None:
                continue
            if not self.in_current(key) and self.items_available.get(key) is not None:
                can_add = True
                break
        self.window.ui.nodes["models.importer.add"].setEnabled(can_add)

    def change_current(self):
        """On change current model selection"""
        if self.provider not in self.items_current:
            self.items_current[self.provider] = {}
        view = self.window.ui.nodes["models.importer.current"]
        sel_model = view.selectionModel()
        rows = sel_model.selectedRows() if sel_model else []
        # collect selected model keys by reading stored tooltip role (stable ID)
        selected_ids = []
        for ix in rows:
            key = ix.data(QtCore.Qt.ToolTipRole)
            if not key:
                key = self.get_by_idx(ix.row(), self.items_current[self.provider])
            if key and key not in selected_ids:
                selected_ids.append(key)

        self.selected_current_ids = selected_ids
        if not selected_ids:
            self.selected_current = None
            self.window.ui.nodes["models.importer.remove"].setEnabled(False)
            return

        # keep a single reference for backward compatibility (last selected)
        self.selected_current = selected_ids[-1]

        # enable remove if any of selected can be removed
        can_remove = False
        for key in selected_ids:
            if key is None:
                continue
            if (key in self.items_current[self.provider]) and self.in_available(key):
                can_remove = True
                break
        self.window.ui.nodes["models.importer.remove"].setEnabled(can_remove)

    def in_available(self, model: str) -> bool:
        """
        Check if model is in available list (all)

        :param model: model ID
        :return: True if model is in available list, False otherwise
        """
        if model in self.items_available_all:
            return True
        for key in list(self.items_available_all.keys()):
            m = self.items_available_all[key]
            if m.id == model:
                return True
        return False

    def _selected_keys_from_view(self, node_id: str, items: Dict) -> List[str]:
        """
        Resolve selected model keys from a given list view using stored tooltip role,
        falling back to index->dict mapping when needed.

        :param node_id: ui node id
        :param items: dict used to build the view's model
        :return: list of selected keys
        """
        keys: List[str] = []
        try:
            view = self.window.ui.nodes[node_id]
            sel_model = view.selectionModel()
            rows = sel_model.selectedRows() if sel_model else []
            for ix in rows:
                key = ix.data(QtCore.Qt.ToolTipRole)
                if not key:
                    key = self.get_by_idx(ix.row(), items)
                if key and key not in keys:
                    keys.append(key)
        except Exception:
            pass
        return keys

    def add(self):
        """Add model(s) to current list"""
        if self.provider not in self.items_current:
            self.items_current[self.provider] = {}

        # collect multi-selection; fallback to single selection
        keys = self._selected_keys_from_view(
            'models.importer.available',
            self.items_available,
        )
        if not keys and self.selected_available is not None:
            keys = [self.selected_available]

        if not keys:
            self.set_status(trans('models.importer.error.add.no_model'))
            return

        any_added = False
        for key in list(keys):
            if key is None:
                continue
            if self.in_current(key):
                continue
            model = self.items_available.get(key)
            if model is None:
                continue
            self.items_current[self.provider][key] = model
            if key not in self.pending:
                self.pending[key] = model
            if key in self.removed:
                del self.removed[key]
            if not self.all and key in self.items_available:
                del self.items_available[key]
            any_added = True

        if not any_added:
            self.set_status(trans('models.importer.error.add.not_exists'))
            return

        self.refresh()

    def remove(self):
        """Remove model(s) from current list"""
        if self.provider not in self.items_current:
            self.items_current[self.provider] = {}

        # collect multi-selection; fallback to single selection
        keys = self._selected_keys_from_view(
            'models.importer.current',
            self.items_current[self.provider],
        )
        if not keys and self.selected_current is not None:
            keys = [self.selected_current]

        if not keys:
            self.set_status(trans('models.importer.error.remove.no_model'))
            return

        any_removed = False
        for key in list(keys):
            if key is None:
                continue
            if not self.in_current(key):
                continue
            model = self.items_current[self.provider].get(key)
            if model is None:
                continue
            # return to available list
            self.items_available[key] = model
            if key not in self.removed:
                self.removed[key] = model
            if key in self.items_current[self.provider]:
                del self.items_current[self.provider][key]
            if key in self.pending:
                del self.pending[key]
            any_removed = True

        if not any_removed:
            self.set_status(trans('models.importer.error.remove.not_exists'))
            return

        self.refresh()

    def setup(self):
        """Set up importer"""
        idx = None
        option = self.get_providers_option()
        self.window.model_importer.setup(idx)  # widget dialog setup
        self.window.ui.config["models.importer"]["provider"].set_keys(option["keys"])
        self.window.ui.add_hook("update.models.importer.provider", self.hook_update)

    def toggle_editor(self):
        """Toggle models importer dialog"""
        if self.dialog:
            self.close()
        else:
            self.open()

    def open(self, force: bool = False):
        """
        Open models editor dialog

        :param force: force open dialog
        """
        if not self.initialized:
            self.setup()
            self.initialized = True
        if not self.dialog or force:
            self.provider = "_"  # default provider
            self.pending = {}
            self.removed = {}
            self.items_current = {}
            self.items_available = {}
            self.items_available_all = {}
            option = self.get_providers_option()
            self.window.controller.config.apply_value(
                parent_id="models.importer",
                key="provider",
                option=option,
                value="_",
            )
            self.init()
            self.window.ui.dialogs.open(
                "models.importer",
                width=self.width,
                height=self.height,
            )
            self.dialog = True

    def close(self):
        """Close models importer dialog"""
        if self.dialog:
            self.window.ui.dialogs.close('models.importer')
            self.dialog = False

    def cancel(self):
        """Cancel models importer dialog"""
        self.close()

    def init(
            self,
            reload: bool = False,
            on_change: bool = False
    ):
        """
        Initialize importer

        :param reload: force re-initialize
        :param on_change: on change provider
        """
        if (self.initialized
                and self.window.ui.nodes["models.importer.available.all"].isChecked()):
            self.all = True
        self.update_title()
        if not reload:
            self.items_available = self.get_available()
        if not on_change:
            self.items_current[self.provider] = self.get_current()
        else:
            # on change provider, only if not loaded
            if self.provider not in self.items_current:
                self.items_current[self.provider] = {}
                self.items_current[self.provider] = self.get_current()
        self.refresh(reload=reload)

    def update_title(self):
        """Update dialog title with current provider"""
        if self.provider == "ollama":
            current = "http://localhost:11434"
            if 'OLLAMA_API_BASE' in os.environ:
                current = os.environ['OLLAMA_API_BASE']
        elif self.provider == "_":
            current = trans("models.importer.current.default")
            self.set_status("")
        else:
            current = self.window.core.llm.get_provider_name(self.provider)
        self.window.ui.nodes["models.importer.url"].setText(current)

    def toggle_all(self, all: bool):
        """
        Toggle all models visibility

        :param all: show all models, not only available for import
        """
        self.all = all
        self.refresh(reload=True)

    def set_status(self, status: str):
        """
        Set status message

        :param status: status message
        """
        if self.initialized:
            self.window.ui.nodes["models.importer.status"].setText(status)

    def get_current(self) -> Dict:
        """
        Get current ollama models

        :return: PyGPT ollama models dictionary
        """
        if self.provider == "_":
            return {}
        items = copy.deepcopy(self.window.core.models.items)
        for key in list(items.keys()):
            if items[key].provider != self.provider:
                del items[key]
        return items

    def get_available(self) -> Dict:
        """
        Get available models for import

        :return: available models dictionary
        """
        models = {}
        if self.provider == "ollama":
            models = self.get_ollama_available()
        else:
            try:
                models = self.get_provider_available()
            except Exception as e:
                print(e)
                self.set_status(str(e))
        models = dict(sorted(models.items(), key=lambda item: item[1].id.lower()))
        return models

    def get_provider_available(self) -> Dict:
        """
        Get available provider models

        :return: Provider API models dictionary
        """
        models = {}
        if not self.provider or self.provider == "_":
            return models
        llm = self.window.core.llm.get(self.provider)
        if llm is None:
            return models
        models_list = llm.get_models(self.window)
        if not models_list or not isinstance(models_list, list):
            return models
        for model in models_list:
            if "id" not in model:
                continue
            id = model.get('id')
            name = model.get('id')
            if "name" in model:
                name = model.get('name')
            m, _ = self.window.core.models.create_empty(append=False)
            m.id = id
            m.name = name
            m.mode = [
                MODE_CHAT,
                MODE_LLAMA_INDEX,
                MODE_AGENT,
                MODE_AGENT_LLAMA,
                MODE_AGENT_OPENAI,
                MODE_EXPERT,
            ]
            m.provider = self.provider
            m.input = ["text"]
            m.output = ["text"]
            llama_id = id
            if self.provider == "google":
                llama_id = "models/" + id
            m.llama_index['args'] = [
                {
                    'name': 'model',
                    'value': llama_id,
                    'type': 'str'
                }
            ]
            m.imported = True
            m.ctx = 128000 # default context size
            key = m.id

            # prepare args and env config by provider
            if self.provider == "anthropic":
                m.tool_calls = True
                m.llama_index['env'] = [
                    {
                        'name': 'ANTHROPIC_API_KEY',
                        'value': '{api_key_anthropic}',
                        'type': 'str'
                    }
                ]
            elif self.provider == "deepseek_api":
                m.tool_calls = True
                m.llama_index['args'].append(
                    {
                        'name': 'api_key',
                        'value': '{api_key_deepseek}',
                        'type': 'str'
                    }
                )
                m.llama_index['env'] = [
                    {
                        'name': 'DEEPSEEK_API_KEY',
                        'value': '{api_key_deepseek}',
                        'type': 'str'
                    }
                ]
            elif self.provider == "google":
                m.tool_calls = True
                m.llama_index['env'] = [
                    {
                        'name': 'GOOGLE_API_KEY',
                        'value': '{api_key_google}',
                        'type': 'str'
                    }
                ]
            elif self.provider in ["openai", "azure_openai"]:
                m.tool_calls = True
                m.llama_index['env'] = [
                    {
                        'name': 'OPENAI_API_KEY',
                        'value': '{api_key}',
                        'type': 'str'
                    },
                    {
                        'name': 'OPENAI_API_BASE',
                        'value': '{api_endpoint}',
                        'type': 'str'
                    },
                    {
                        'name': 'AZURE_OPENAI_ENDPOINT',
                        'value': '{api_azure_endpoint}',
                        'type': 'str'
                    },
                    {
                        'name': 'OPENAI_API_VERSION',
                        'value': '{api_azure_version}',
                        'type': 'str'
                    }
                ]
            elif self.provider == "perplexity":
                m.tool_calls = True
                m.llama_index['env'] = [
                    {
                        'name': 'OPENAI_API_KEY',
                        'value': '{api_key_perplexity}',
                        'type': 'str'
                    },
                    {
                        'name': 'OPENAI_API_BASE',
                        'value': '{api_endpoint_perplexity}',
                        'type': 'str'
                    }
                ]
            elif self.provider == "mistral_ai":
                m.tool_calls = True
                m.llama_index['args'].append(
                    {
                        'name': 'api_key',
                        'value': '{api_key_mistral}',
                        'type': 'str'
                    }
                )
            elif self.provider == "local_ai":
                m.tool_calls = True
                m.llama_index['env'] = [
                    {
                        'name': 'OPENAI_API_KEY',
                        'value': '{api_key}',
                        'type': 'str'
                    },
                    {
                        'name': 'OPENAI_API_BASE',
                        'value': '{api_endpoint}',
                        'type': 'str'
                    }
                ]
            elif self.provider == "open_router":
                m.tool_calls = True
                m.llama_index['env'] = [
                    {
                        'name': 'OPENAI_API_KEY',
                        'value': '{api_key_open_router}',
                        'type': 'str'
                    },
                    {
                        'name': 'OPENAI_API_BASE',
                        'value': '{api_endpoint_open_router}',
                        'type': 'str'
                    }
                ]
            elif self.provider == "x_ai":
                m.tool_calls = True
                m.llama_index['env'] = [
                    {
                        'name': 'OPENAI_API_KEY',
                        'value': '{api_key_xai}',
                        'type': 'str'
                    },
                    {
                        'name': 'OPENAI_API_BASE',
                        'value': '{api_endpoint_xai}',
                        'type': 'str'
                    }
                ]
            models[key] = m
        provider_name = self.window.core.llm.get_provider_name(self.provider)
        self.set_status(trans('models.importer.loaded').replace("{provider}", provider_name))
        return models

    def get_ollama_available(self) -> Dict:
        """
        Get available ollama models

        :return: Ollama API models dictionary
        """
        models = {}
        status = self.window.core.models.ollama.get_status()
        if not status['status']:
            self.set_status(trans('models.importer.error.no_connection'))
            return models
        else:
            ollama_models = status.get('models', [])
            if not ollama_models:
                self.set_status(trans('models.importer.error.no_models'))
                return models
            else:
                for model in ollama_models:
                    name = model.get('name')
                    m, _ = self.window.core.models.create_empty(append=False)
                    m.id = name
                    m.name = name
                    m.mode = [
                        MODE_CHAT,
                        MODE_LLAMA_INDEX,
                        MODE_AGENT,
                        MODE_AGENT_LLAMA,
                        MODE_AGENT_OPENAI,
                        MODE_EXPERT,
                    ]
                    m.provider = 'ollama'
                    m.input = ["text"]
                    m.output = ["text"]
                    m.llama_index['args'] = [
                        {
                            'name': 'model',
                            'value': name,
                            'type': 'str'
                        }
                    ]
                    m.imported = True
                    m.ctx = 32000  # default
                    key = m.id
                    models[key] = m
                self.set_status(trans('models.importer.loaded').replace("{provider}", "Ollama"))
        return models

    def from_pending(self):
        """Move pending models to base list"""
        added = False
        base_models = self.window.core.models.items
        for key in list(self.pending.keys()):
            if key not in base_models:
                base_models[key] = copy.deepcopy(self.pending[key])
                base_models[key].imported = True
                added = True
            else:
                # add provider suffix - to key
                new_key = f"{key}-{self.provider}"
                if new_key not in base_models:
                    base_models[new_key] = copy.deepcopy(self.pending[key])
                    base_models[new_key].imported = True
                    added = True
        for key in list(self.removed.keys()):
            if key in base_models:
                del base_models[key]
                added = True
        self.pending = {}
        self.removed = {}
        if added:
            self.window.core.models.sort_items()
            self.window.core.models.save()
            self.set_status(trans('models.importer.status.imported'))

    def save(self, persist: bool = True):
        """
        Save models

        :param persist: persist to file and close dialog
        """
        self.from_pending()
        self.window.controller.model.init_list()
        self.window.controller.model.update()
        self.close()

        # dispatch on update event
        event = Event(Event.MODELS_CHANGED)
        self.window.dispatch(event, all=True)

    def refresh(self, reload: bool = False):
        """
        Reload items

        :param reload: reload available models
        """
        if self.provider == "_":
            self.set_status("")

        if reload:
            if self.provider == "_":
                self.set_status("")
            else:
                self.set_status(trans("models.importer.status.wait"))
            QApplication.processEvents()
            self.items_available = self.get_available()
            self.items_available_all = copy.deepcopy(self.items_available)

        # remove from available if already in current
        if not self.all:
            for key in list(self.items_available.keys()):
                if self.in_current(key):
                    del self.items_available[key]

        self.window.ui.nodes['models.importer.editor'].update_available(self.items_available)
        self.window.ui.nodes['models.importer.editor'].update_current(self.items_current[self.provider])

    def get_by_idx(self, idx: int, items: Dict) -> Optional[str]:
        """
        Get model key by list index

        :param idx: list index
        :param items: items dictionary
        :return: model key
        """
        model_idx = 0
        for id in self.get_ids(items):
            if model_idx == idx:
                return id
            model_idx += 1
        return None

    def get_ids(self, items: Dict) -> List[str]:
        """
        Return models ids

        :return: model ids list
        """
        return list(items.keys())

    def get_providers_option(self) -> Dict:
        """
        Get available LLM providers option

        :return: Dict with keys and values
        """
        excluded = [
            #"azure_openai",
            "huggingface_api",
            #"mistral_ai",
            "local_ai",
            "perplexity",
        ]
        choices = self.window.core.llm.get_choices()
        data = []
        data.append({"_": trans("models.importer.list.select")})
        for id in choices:
            if id in excluded:
                continue
            data.append({id: choices[id]})

        option = {
            "type": "combo",
            "label": "Provider",
            "keys": data,
        }
        return option

    def hook_update(
            self,
            key: str,
            value: Any,
            caller,
            *args,
            **kwargs
    ):
        """
        Hook: on settings update in real-time (prompt)

        :param key: setting key
        :param value: setting value
        :param caller: caller id
        :param args: additional arguments
        :param kwargs: additional keyword arguments
        """
        # show/hide extra options
        if key == "provider":
            self.provider = value
            self.init(reload=True, on_change=True)
            if value == "_":
                self.set_status("")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.02 20:00:00                  #
# ================================================== #

from typing import Optional, Dict, Any, List

from pygpt_net.core.types import (
    MODE_COMPLETION,
)
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem


class Completion:
    def __init__(self, window=None):
        """
        Completion wrapper

        :param window: Window instance
        """
        self.window = window
        self.input_tokens = 0

    def send(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ):
        """
        Call OpenAI API for completion

        :param context: Bridge context
        :param extra: Extra arguments
        :return: response or stream chunks
        """
        prompt = context.prompt
        stream = context.stream
        max_tokens = int(context.max_tokens or 0)
        system_prompt = context.system_prompt
        model = context.model
        model_id = model.id

        ctx = context.ctx
        if ctx is None:
            ctx = CtxItem()  # create empty context
        user_name = ctx.input_name  # from ctx
        ai_name = ctx.output_name  # from ctx

        # build prompt message
        message = self.build(
            prompt=prompt,
            system_prompt=system_prompt,
            model=model,
            history=context.history,
            ai_name=ai_name,
            user_name=user_name,
        )

        # check if max tokens not exceeded
        if model.ctx > 0:
            available_tokens = model.ctx - self.input_tokens
            if max_tokens > 0:
                if available_tokens < max_tokens:
                    max_tokens = available_tokens

        # prepare stop word if user_name is set
        stop = ""
        if user_name is not None and user_name != '':
            stop = [user_name + ':']

        client = self.window.core.api.openai.get_client()

        # fix for deprecated OpenAI davinci models
        if model_id.startswith('text-davinci'):
            model_id = 'gpt-3.5-turbo-instruct'

        # extra API kwargs
        response_kwargs = {}
        if max_tokens > 0:
            response_kwargs['max_tokens'] = max_tokens

        response = client.completions.create(
            prompt=message,
            model=model_id,
            temperature=self.window.core.config.get('temperature'),
            top_p=self.window.core.config.get('top_p'),
            frequency_penalty=self.window.core.config.get('frequency_penalty'),
            presence_penalty=self.window.core.config.get('presence_penalty'),
            stop=stop,
            stream=stream,
            **response_kwargs
        )
        return response

    def build(
            self,
            prompt: str,
            system_prompt: str,
            model: ModelItem,
            history: Optional[List[CtxItem]] = None,
            ai_name: Optional[str] = None,
            user_name: Optional[str] = None,
    ) -> str:
        """
        Build completion string

        :param prompt: user prompt
        :param system_prompt: system prompt
        :param model: model item
        :param history: history
        :param ai_name: AI name
        :param user_name: username
        :return: message string (parsed with context)
        """
        message = ""

        # tokens config
        used_tokens = self.window.core.tokens.from_user(
            prompt,
            system_prompt,
        )
        max_ctx_tokens = self.window.core.config.get('max_total_tokens')

        # fit to max model ctx tokens
        if max_ctx_tokens > model.ctx and model.ctx > 0:
            max_ctx_tokens = model.ctx

        # input tokens: reset
        self.reset_tokens()

        if system_prompt is not None and system_prompt != "":
            message += system_prompt

        if self.window.core.config.get('use_context'):
            items = self.window.core.ctx.get_history(
                history,
                model.id,
                MODE_COMPLETION,
                used_tokens,
                max_ctx_tokens,
            )
            for item in items:
                if item.input_name is not None \
                        and item.output_name is not None \
                        and item.input_name != "" \
                        and item.output_name != "":
                    if item.final_input is not None and item.final_input != "":
                        message += "\n" + item.input_name + ": " + item.final_input
                    if item.final_output is not None and item.final_output != "":
                        message += "\n" + item.output_name + ": " + item.final_output
                else:
                    if item.final_input is not None and item.final_input != "":
                        message += "\n" + item.final_input
                    if item.final_output is not None and item.final_output != "":
                        message += "\n" + item.final_output

        # append names
        if user_name is not None \
                and ai_name is not None \
                and user_name != "" \
                and ai_name != "":
            message += "\n" + user_name + ": " + str(prompt)
            message += "\n" + ai_name + ":"
        else:
            message += "\n" + str(prompt)

        # input tokens: update
        self.input_tokens += self.window.core.tokens.from_text(
            message,
            model.id,
        )
        return message

    def reset_tokens(self):
        """Reset input tokens counter"""
        self.input_tokens = 0

    def get_used_tokens(self) -> int:
        """
        Get input tokens counter

        :return: input tokens
        """
        return self.input_tokens



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.31 23:00:00                  #
# ================================================== #

from pygpt_net.core.events import RealtimeEvent

def build_rt_input_delta_event(
        rate: int,
        channels: int,
        data: bytes,
        final: bool
) -> RealtimeEvent:
    """
    Build RT_INPUT_AUDIO_DELTA event with a provider-agnostic payload.

    :param rate: Sample rate (e.g., 16000)
    :param channels: Number of channels (1 for mono, 2 for stereo)
    :param data: Audio data bytes
    :param final: Whether this is the final chunk
    :return: RealtimeEvent instance
    """
    return RealtimeEvent(
        RealtimeEvent.RT_INPUT_AUDIO_DELTA,
        {
            "payload": {
                "data": data or b"",
                "mime": "audio/pcm",
                "rate": int(rate),
                "channels": int(channels),
                "final": bool(final),
            }
        }
    )

def build_output_volume_event(value: int) -> RealtimeEvent:
    """
    Build RT_OUTPUT_AUDIO_VOLUME_CHANGED event.

    :param value: Volume level (0-100)
    :return: RealtimeEvent instance
    """
    return RealtimeEvent(
        RealtimeEvent.RT_OUTPUT_AUDIO_VOLUME_CHANGED,
        {"volume": int(value)}
    )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.28 00:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QLabel, QHBoxLayout, QWidget

from datetime import datetime
from pygpt_net.utils import trans


class BottomStatus:
    def __init__(self, window=None):
        self.window = window
        self.timer = QLabel(parent=self.window)
        self.timer.setObjectName("StatusBarTimer")
        self.msg = QLabel(parent=self.window)
        self.msg.setObjectName("StatusBarMessage")
        self.set_text(trans('status.started'))

    def set_text(self, text):
        """Set status text"""
        self.msg.setText(text)
        if text:
            now = datetime.now()
            self.timer.setText(now.strftime("%H:%M"))
        else:
            self.timer.setText("")

    def setText(self, text):
        """Fallback for set_text method"""
        self.set_text(text)

    def text(self) -> str:
        """Get status text"""
        return self.msg.text()

    def setup(self):
        """Setup status bar widget"""
        self.timer.setText("00:00")
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(5)
        layout.addWidget(self.timer)
        layout.addWidget(self.msg)
        layout.addStretch()
        widget = QWidget(self.window)
        widget.setLayout(layout)
        return widget


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.03 14:00:00                  #
# ================================================== #

from typing import Optional, Tuple

from pygpt_net.core.bridge.worker import BridgeSignals
from pygpt_net.item.ctx import CtxItem


class BaseRunner:
    def __init__(self, window=None):
        """
        Agent runner base class

        :param window: Window instance
        """
        self.window = window

    def add_ctx(
            self,
            from_ctx: CtxItem,
            with_tool_outputs: bool = False
    ) -> CtxItem:
        """
        Prepare response context item

        :param from_ctx: CtxItem (parent, source)
        :param with_tool_outputs: True if tool outputs should be copied
        :return: CtxItem with copied data from parent context item
        """
        return self.window.core.agents.runner.helpers.add_ctx(
            from_ctx=from_ctx,
            with_tool_outputs=with_tool_outputs
        )

    def add_next_ctx(
            self,
            from_ctx: CtxItem,
    ) -> CtxItem:
        """
        Prepare next context item (used for new context in the cycle)

        :param from_ctx: CtxItem (parent, source)
        :return: CtxItem with copied data from parent context item
        """
        return self.window.core.agents.runner.helpers.add_next_ctx(
            from_ctx=from_ctx
        )

    def send_stream(
            self,
            ctx: CtxItem,
            signals: BridgeSignals,
            begin: bool=False
    ):
        """
        Send stream chunk to chat window (BridgeSignals)

        :param ctx: CtxItem
        :param signals: BridgeSignals
        :param begin: True if it is the beginning of the text
        """
        self.window.core.agents.runner.helpers.send_stream(
            ctx=ctx,
            signals=signals,
            begin=begin
        )

    def end_stream(self, ctx: CtxItem, signals: BridgeSignals):
        """
        End of stream in chat window (BridgeSignals)

        :param ctx: CtxItem
        :param signals: BridgeSignals
        """
        self.window.core.agents.runner.helpers.end_stream(
            ctx=ctx,
            signals=signals
        )

    def next_stream(self, ctx: CtxItem, signals: BridgeSignals):
        """
        End of stream in chat window (BridgeSignals)

        :param ctx: CtxItem
        :param signals: BridgeSignals
        """
        self.window.core.agents.runner.helpers.next_stream(
            ctx=ctx,
            signals=signals
        )

    def send_response(
            self,
            ctx: CtxItem,
            signals: BridgeSignals,
            event_name: str,
            **kwargs
    ):
        """
        Send async response to chat window (BridgeSignals)

        :param ctx: CtxItem
        :param signals: BridgeSignals
        :param event_name: kernel event
        :param kwargs: extra data
        """
        self.window.core.agents.runner.helpers.send_response(
            ctx=ctx,
            signals=signals,
            event_name=event_name,
            **kwargs
        )

    def set_busy(
            self,
            signals: BridgeSignals,
            **kwargs
    ):
        """
        Set busy status

        :param signals: BridgeSignals
        :param kwargs: extra data
        """
        self.window.core.agents.runner.helpers.set_busy(
            signals=signals,
            **kwargs
        )

    def set_idle(
            self,
            signals: BridgeSignals,
            **kwargs
    ):
        """
        Set idle status

        :param signals: BridgeSignals
        :param kwargs: extra data
        """
        self.window.core.agents.runner.helpers.set_idle(
            signals=signals,
            **kwargs
        )

    def set_status(
            self,
            signals: BridgeSignals,
            msg: str
    ):
        """
        Set busy status

        :param signals: BridgeSignals
        :param msg: status message
        """
        self.window.core.agents.runner.helpers.set_status(
            signals=signals,
            msg=msg
        )

    def prepare_input(self, prompt: str) -> str:
        """
        Prepare input context

        :param prompt: input text
        """
        return self.window.core.agents.runner.helpers.prepare_input(
            prompt=prompt
        )

    def is_stopped(self) -> bool:
        """
        Check if run is stopped

        :return: True if stopped
        """
        return self.window.core.agents.runner.helpers.is_stopped()

    def set_error(self, error: Exception):
        """
        Set last error

        :param error: Exception to set
        """
        self.window.core.agents.runner.helpers.set_error(
            error=error
        )

    def get_error(self) -> Optional[Exception]:
        """
        Get last error

        :return: last exception or None if no error
        """
        return self.window.core.agents.runner.helpers.get_error()

    def extract_final_response(self, input_text: str) -> Tuple[str, str]:
        """
        Extract final response from input text.

        :param input_text: str
        :return: thought and answer
        """
        return self.window.core.agents.runner.helpers.extract_final_response(
            input_text=input_text
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.24 02:00:00                  #
# ================================================== #

import json
from typing import List, Dict, Any

from agents import (
    FunctionTool as OpenAIFunctionTool,
    RunContextWrapper,
)
from llama_index.core.chat_engine.types import AgentChatResponse
from llama_index.core.tools import BaseTool, FunctionTool, QueryEngineTool, ToolMetadata

from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.events import Event
from pygpt_net.core.types import (
    TOOL_QUERY_ENGINE_NAME,
    TOOL_QUERY_ENGINE_DESCRIPTION,
    TOOL_QUERY_ENGINE_SPEC,
)
from pygpt_net.item.ctx import CtxItem


class Tools:

    def __init__(self, window=None):
        """
        Agent tools

        :param window: Window instance
        """
        self.window = window
        self.cmd_blacklist = []
        self.verbose = False
        self.code_execute_fn = CodeExecutor(window)
        self.last_tool_output = None
        self.agent_idx = None  # agent index, used for query engine tool
        self.context = None  # BridgeContext instance, used for tool execution

    def prepare(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            verbose: bool = False,
            force: bool = False
    ) -> List[BaseTool]:
        """
        Prepare tools for agent

        :param context: BridgeContext
        :param extra: extra data
        :param verbose: verbose mode
        :param force: force to get functions even if not needed
        :return: list of tools
        """
        self.verbose = verbose
        tools = []

        # add functions from plugins
        plugin_functions = self.get_plugin_functions(context.ctx, verbose=verbose, force=force)
        tools.extend(plugin_functions)

        # add query engine tool if idx is provided
        query_engine_tools = self.get_retriever_tool(
            context=context,
            extra=extra,
            verbose=verbose,
        )
        if query_engine_tools:
            tools.extend(query_engine_tools)
        return tools

    def get_retriever_tool(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            verbose: bool = False
    ) -> List[BaseTool]:
        """
        Prepare tools for agent

        :param context: BridgeContext
        :param extra: extra data
        :param verbose: verbose mode
        :return: list of tools
        """
        tool = None

        # add query engine tool if idx is provided
        idx = extra.get("agent_idx", None)
        if self.window.core.idx.is_valid(idx):
            llm, embed_model = self.window.core.idx.llm.get_service_context(model=context.model)
            index = self.window.core.idx.storage.get(idx, llm, embed_model)  # get index
            if index is not None:
                query_engine = index.as_query_engine(similarity_top_k=3)
                tool = [
                    QueryEngineTool(
                        query_engine=query_engine,
                        metadata=ToolMetadata(
                            name=TOOL_QUERY_ENGINE_NAME,
                            description=TOOL_QUERY_ENGINE_DESCRIPTION,
                        ),
                    ),
                ]
        return tool

    def get_openai_retriever_tool(
            self,
            idx: str,
            verbose: bool = False
    ) -> OpenAIFunctionTool:
        """
        Prepare OpenAI retriever tool for agent

        :param idx: index name
        :param verbose: verbose mode
        :return: OpenAIFunctionTool instance
        """
        async def run_function(run_ctx: RunContextWrapper[Any], args: str) -> str:
            name = run_ctx.tool_name
            print(f"[Plugin] Tool call: {name} with args: {args}")
            cmd = {
                "cmd": name,
                "params": json.loads(args)  # args should be a JSON string
            }
            return self.tool_exec(name, cmd["params"])

        schema = {"type": "object", "properties": {
            "query": {
                "type": "string",
                "description": "The query string to search in the index."
            }
        }, "additionalProperties": False}
        description = TOOL_QUERY_ENGINE_DESCRIPTION + f" Index: {idx}"
        return OpenAIFunctionTool(
            name=TOOL_QUERY_ENGINE_NAME,
            description=description,
            params_json_schema=schema,
            on_invoke_tool=run_function,
        )

    def get_plugin_functions(
            self,
            ctx: CtxItem,
            verbose: bool = False,
            force: bool = False
    ) -> List[BaseTool]:
        """
        Parse plugin functions

        :param ctx: CtxItem
        :param verbose: verbose mode
        :param force: force to get functions even if not needed
        :return: List of BaseTool instances
        """
        tools = []
        functions = self.window.core.command.get_functions(force=force)
        for item in functions:
            try:
                name = item['name']
                if name in self.cmd_blacklist:
                    continue  # skip blacklisted commands

                description = item['desc']
                schema = json.loads(item['params'])  # from JSON to dict

                def make_func(name, description):
                    def func(**kwargs):
                        self.log(f"[Plugin] Tool call: {name} {kwargs}")
                        cmd = {
                            "cmd": name,
                            "params": kwargs,
                        }
                        response = self.window.controller.plugins.apply_cmds_all(
                            ctx,  # current ctx
                            [cmd],  # commands
                        )
                        return str(response)  # return response as string

                    func.__name__ = name
                    func.__doc__ = description
                    return func

                func = make_func(name, description)
                metadata = PluginToolMetadata(
                    name=name,
                    description=description,
                )
                metadata.schema = schema
                tool = FunctionTool(
                    fn=func,
                    metadata=metadata,
                )
                tools.append(tool)
            except Exception as e:
                self.window.core.debug.log(e)
        return tools

    def get_function_tools(
            self,
            ctx: CtxItem,
            verbose: bool = False,
            force: bool = False
    ) -> List[OpenAIFunctionTool]:
        """
        Parse plugin functions and return as OpenAI FunctionTool instances

        :param ctx: CtxItem
        :param verbose: verbose mode
        :param force: force to get functions even if not needed
        :return: List of OpenAIFunctionTool instances
        """
        tools = []
        functions = self.window.core.command.get_functions(force=force)
        blacklist = []
        for item in functions:
            try:
                name = item['name']
                if name in self.cmd_blacklist or name in blacklist:
                    continue
                description = item['desc']

                async def run_function(run_ctx: RunContextWrapper[Any], args: str) -> str:
                    name = run_ctx.tool_name
                    print(f"[Plugin] Tool call: {name} with args: {args}")
                    cmd = {
                        "cmd": name,
                        "params": json.loads(args)  # args should be a JSON string
                    }
                    return self.window.controller.plugins.apply_cmds_all(ctx, [cmd])

                schema = json.loads(item['params'])  # from JSON to dict
                extra = ""
                # fix schema for OpenAI FunctionTool
                if "properties" in schema:
                    for property_name, property_value in schema["properties"].items():
                        if "enum" in property_value:
                            """
                            extra += ", Enum values: "
                            extra += ", ".join(
                                [str(enum_value) for enum_value in property_value["enum"]]
                            )
                            """
                            del property_value["enum"]  # remove enum for OpenAI FunctionTool
                        if property_value["type"] == "object":
                            if "properties" not in property_value:
                                property_value["properties"] = {}
                            if "required" not in property_value:
                                property_value["required"] = []
                            if "additionalProperties" not in property_value:
                                property_value["additionalProperties"] = False
                schema["additionalProperties"] = False
                description += extra
                tool = OpenAIFunctionTool(
                    name=name,
                    description=description,
                    params_json_schema=schema,
                    on_invoke_tool=run_function,
                )
                tools.append(tool)
            except Exception as e:
                self.window.core.debug.log(e)

        # append query engine tool if idx is provided
        if self.window.core.idx.is_valid(self.agent_idx):
            tools.append(self.get_openai_retriever_tool(self.agent_idx))

        return tools

    def get_plugin_tools(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            verbose: bool = False,
            force: bool = False
    ) -> Dict[str, BaseTool]:
        """
        Parse plugin functions

        :param context: BridgeContext
        :param extra: extra data
        :param verbose: verbose mode
        :param force: force to get functions even if not needed
        :return: Dictionary of tool names and BaseTool instances
        """
        tools = {}
        functions = self.window.core.command.get_functions(force=force)
        for item in functions:
            try:
                name = item['name']
                if name in self.cmd_blacklist:
                    continue  # skip blacklisted commands

                description = item['desc']

                def make_func(name, description):
                    def func(**kwargs):
                        self.log(f"[Plugin] Tool call: {name} {kwargs}")
                        cmd = {
                            "cmd": name,
                            "params": kwargs,
                        }
                        response = self.window.controller.plugins.apply_cmds_all(
                            context.ctx,  # current ctx
                            [cmd],  # commands
                        )
                        return str(response)  # return response as string

                    func.__name__ = name
                    func.__doc__ = description
                    return func

                func = make_func(name, description)
                tools[name] = func
            except Exception as e:
                self.window.core.debug.log(e)

        # add query engine tool if idx is provided
        if self.window.core.idx.is_valid(self.agent_idx):
            extra = {
                "agent_idx": self.agent_idx,  # agent index for query engine tool
            }
            query_engine_tools = self.get_retriever_tool(
                context=context,
                extra=extra,
                verbose=verbose,
            )
            if query_engine_tools:
                tools["query_engine"] = query_engine_tools[0]  # add query engine tool
        return tools


    def get_plugin_specs(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            verbose: bool = False,
            force: bool = False
    ) -> List[str]:
        """
        Parse plugin functions

        :param context: BridgeContext
        :param extra: extra data
        :param verbose: verbose mode
        :param force: force to get functions even if not needed
        :return: List of tool specifications as strings
        """
        specs = []
        functions = self.window.core.command.get_functions(force=force)

        # add query engine tool spec if idx is provided
        if self.window.core.idx.is_valid(self.agent_idx):
            specs.append(TOOL_QUERY_ENGINE_SPEC)

        for func in functions:
            try:
                name = func['name']
                if name in self.cmd_blacklist:
                    continue  # skip blacklisted commands
                description = func['desc']
                schema = json.loads(func['params'])  # from JSON to dict
                specs.append(
                    f"**{name}**: {description}, available params: {schema.get('properties', {})}, required: {schema.get('required', [])}\n"
                )
            except Exception as e:
                self.window.core.debug.log(e)
        return specs

    def tool_exec(self, cmd: str, params: Dict[str, Any]) -> str:
        """
        Execute tool command

        :param cmd: command name
        :param params: command parameters
        :return: command output
        """
        print(f"[Plugin] Tool call: {cmd}, {params}")

        # special case for query engine tool
        if cmd == TOOL_QUERY_ENGINE_NAME:
            if "query" not in params:
                return "Query parameter is required for query_engine tool."
            if self.context is None:
                return "Context is not set for query_engine tool."
            if not self.window.core.idx.is_valid(self.agent_idx):
                return "Agent index is not set for query_engine tool."
            llm, embed_model = self.window.core.idx.llm.get_service_context(model=self.context.model)
            index = self.window.core.idx.storage.get(self.agent_idx, llm, embed_model)  # get index
            if index is not None:
                query_engine = index.as_query_engine(similarity_top_k=3)
                response = query_engine.query(params["query"])
                print(f"[Plugin] Query engine response: {response}")
                self.log(f"[Plugin] Query engine response: {response}")
                return str(response)
            else:
                return "Index not found for query_engine tool."

        # rest of the plugin commands
        cmd = {
            "cmd": cmd,
            "params": params,
        }
        ctx = CtxItem()
        ctx.extra["agent_input"] = True  # mark as user input
        ctx.agent_call = True  # disables reply from plugin commands
        response = self.window.controller.plugins.apply_cmds_all(
            ctx,  # current ctx
            [cmd],  # commands
        )
        return response

    def export_sources(
            self,
            response: AgentChatResponse
    ) -> List[dict]:
        """
        Export sources from response

        :param response: response
        :return: list of sources
        """
        data = []
        for output in response.sources:
            item = {
                "ToolOutput": {
                    "content": str(output.content),
                    "tool_name": str(output.tool_name),
                    "raw_input": str(output.raw_input),
                    "raw_output": str(output.raw_output),
                }
            }
            data.append(item)
        return data

    def get_last_tool_output(self) -> dict:
        """
        Get last tool output

        :return: last tool output
        """
        if self.last_tool_output is None:
            return {}
        return self.last_tool_output

    def has_last_tool_output(self) -> bool:
        """
        Check if there is a last tool output

        :return: True if last tool output exists, False otherwise
        """
        return self.last_tool_output is not None

    def clear_last_tool_output(self):
        """Clear last tool output"""
        self.last_tool_output = None

    def append_tool_outputs(self, ctx: CtxItem, clear: bool = True):
        """
        Append tool outputs to context

        :param ctx: CtxItem
        :param clear: clear last tool output after appending
        """
        if self.has_last_tool_output():
            outputs = [self.get_last_tool_output()]
            ctx.extra["tool_output"] = outputs
            if outputs is not None:
                response = ""
                for output in outputs:
                    if ("code" in output and "output" in output["code"] and
                            "content" in output["code"]["output"]):
                        response += str(output["code"]["output"]["content"])
                self.window.core.filesystem.parser.extract_data_files(ctx, response) # img, files
            if clear:
                self.clear_last_tool_output()  # clear after use

    def extract_tool_outputs(self, ctx: CtxItem, clear: bool = True):
        """
        Append tool outputs to context

        :param ctx: CtxItem
        :param clear: clear last tool output after appending
        """
        if self.has_last_tool_output():
            outputs = [self.get_last_tool_output()]
            if outputs is not None:
                response = ""
                for output in outputs:
                    if ("code" in output and "output" in output["code"] and
                            "content" in output["code"]["output"]):
                        response += str(output["code"]["output"]["content"])
                self.window.core.filesystem.parser.extract_data_files(ctx, response) # img, files
            if clear:
                self.clear_last_tool_output()  # clear after use

    def set_idx(self, agent_idx: str):
        """
        Set agent index for query engine tool

        :param agent_idx: agent index
        """
        self.agent_idx = agent_idx

    def set_context(self, context: BridgeContext):
        """
        Set context for tool execution

        :param context: BridgeContext instance
        """
        self.context = context

    def log(self, msg: str):
        """
        Log message

        :param msg: message
        """
        if self.verbose:
            print(msg)
            self.window.core.debug.add(msg)

class PluginToolMetadata(ToolMetadata):
    def __init__(self, name: str, description: str):
        super().__init__(name=name, description=description)
        self.schema = None

    def get_parameters_dict(self) -> Dict[str, Any]:
        """
        Get parameters dictionary

        :return: parameters
        """
        parameters = {
            k: v
            for k, v in self.schema.items()
            if k in ["type", "properties", "required", "definitions"]
        }
        return parameters

class CodeExecutor:
    """Code executor for codeAct agent"""

    def __init__(self, window = None):
        """
        Initialize the code executor.

        :param window: Window instance
        """
        self.window = window

    def execute(self, code: str) -> str:
        """
        Execute Python code and capture output and return values.

        :param code: Python code to execute
        :return: Output from the code execution
        """
        if not self.window.core.command.is_cmd():
            return "Tool execution is not enabled. Abort execution and ask user for tool enable."

        self.window.core.agents.tools.last_tool_output = None
        if code == "/restart":
            commands = [
                {
                    "cmd": "ipython_kernel_restart",
                    "params": {},
                    "silent": True,
                    "force": True,
                }
            ]
        else:
            commands = [
                {
                    "cmd": "ipython_execute",
                    "params": {
                        "code": code,
                        "path": ".interpreter.current.py",
                    },
                    "silent": True,
                    "force": True,
                }
            ]
        event = Event(Event.CMD_EXECUTE, {
            'commands': commands,
            'silent': True,
        })
        event.ctx = CtxItem()  # tmp
        event.ctx.async_disabled = True  # disable async for this event
        self.window.controller.command.dispatch_only(event)

        # if restart command was executed, return success message
        if code == "/restart":
            return "IPython kernel restarted successfully."

        response = event.ctx.bag  # tmp response
        output = ""

        # store tool output if available
        if "code" in response:
            if "output" in response["code"]:
                output = response["code"]["output"]["content"]
                tool_output = {
                    "cmd": "ipython_execute",
                    "code": {
                        "input": {
                            "content": response["code"]["input"]["content"],
                            "lang": "python"
                        },
                        "output": {
                            "content": response["code"]["output"]["content"],
                            "lang": "python"
                        }
                    },
                    "plugin": response["plugin"],
                    "result": response["result"]
                }
                self.window.core.agents.tools.last_tool_output = tool_output
        return output



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.12.30 22:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QVBoxLayout, QWidget, QHBoxLayout

from pygpt_net.ui.widget.option.combo import OptionCombo
from pygpt_net.ui.widget.option.input import OptionInput
from pygpt_net.ui.widget.option.toggle_label import ToggleLabel
from pygpt_net.utils import trans


class Image:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window

    def setup(self) -> QWidget:
        """
        Setup image

        :return: QWidget
        :rtype: QWidget
        """
        option = {
            "type": "int",
            "label": "img_variants",
            "min": 1,
            "max": 4,
            "value": 1,
        }

        ui = self.window.ui
        conf_global = ui.config['global']

        container = QWidget(parent=self.window)
        ui.nodes['dalle.options'] = container

        conf_global['img_variants'] = OptionInput(self.window, 'global', 'img_variants', option)
        conf_global['img_variants'].setToolTip(trans("toolbox.img_variants.label"))

        option_resolutions = self.window.core.image.get_resolution_option()
        conf_global['img_resolution'] = OptionCombo(self.window, 'global', 'img_resolution', option_resolutions)
        conf_global['img_resolution'].setMinimumWidth(160)

        conf_global['img.remix'] = ToggleLabel(trans("img.remix"), parent=self.window)
        conf_global['img.remix'].box.setToolTip(trans("img.remix.tooltip"))
        conf_global['img.remix'].box.toggled.connect(self.window.controller.media.toggle_remix_image)

        cols = QHBoxLayout()
        cols.addWidget(conf_global['img_resolution'], 3)
        cols.addWidget(conf_global['img_variants'], 1)
        cols.setContentsMargins(2, 5, 5, 5)

        rows = QVBoxLayout()
        rows.addLayout(cols)
        rows.addWidget(conf_global['img.remix'])
        rows.setContentsMargins(2, 5, 5, 5)

        container.setLayout(rows)
        container.setContentsMargins(2, 0, 0, 10)
        container.setFixedHeight(100)

        return container


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.14 20:00:00                  #
# ================================================== #

import datetime


class IndexesDebug:
    def __init__(self, window=None):
        """
        Indexes debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'indexes'

    def update(self):
        """Update debug window."""
        debug = self.window.core.debug
        idx_core = self.window.core.idx
        idx_controller = self.window.controller.idx
        config = self.window.core.config
        storage = idx_core.storage
        indexing = idx_core.indexing

        debug.begin(self.id)
        debug.add(self.id, 'Current storage:', str(idx_core.get_current_store()))
        debug.add(self.id, 'Current idx:', str(idx_controller.current_idx))
        debug.add(self.id, 'Current mode:', str(idx_controller.current_mode))
        debug.add(self.id, 'Chat mode:', str(config.get("llama.idx.chat.mode")))
        debug.add(self.id, 'Indexes (list):', str(list(config.get("llama.idx.list"))))

        # count items in DB
        db_counter = {
            "ctx": idx_core.get_counters("ctx"),
            "file": idx_core.get_counters("file"),
            "external": idx_core.get_counters("external")
        }

        for store in idx_core.items:
            name = "Items (files): " + store
            indexes_data = {idx: len(idx_core.items[store][idx].items)
                            for idx in idx_core.items[store]}
            if indexes_data:
                debug.add(self.id, name, str(indexes_data))

        # DB items counters
        debug.add(self.id, 'DB items (ctx):', str(db_counter["ctx"]))
        debug.add(self.id, 'DB items (external/web):', str(db_counter["external"]))
        debug.add(self.id, 'DB items (file):', str(db_counter["file"]))

        debug.add(self.id, 'Storage (storages):', str(list(storage.storages.keys())))
        debug.add(self.id, 'Temp (in-memory) indices:', str(storage.count_tmp()))

        # loaders
        debug.add(self.id, 'Offline loaders [files]:', str(sorted(list(indexing.loaders["file"].keys()))))
        debug.add(self.id, 'Offline loaders [web]:', str(sorted(list(indexing.loaders["web"].keys()))))
        debug.add(self.id, 'External instructions [web]:', str(indexing.external_instructions))

        excluded = config.get("llama.idx.excluded.ext").replace(" ", "").split(',')
        debug.add(self.id, 'Excluded (ext):', str(excluded))
        debug.add(self.id, 'Force exclude:', str(config.get("llama.idx.excluded.force")))
        debug.add(self.id, 'Custom metadata:', str(config.get("llama.idx.custom_meta")))

        # ctx
        debug.add(self.id, 'CTX [auto]:', str(config.get("llama.idx.auto")))
        last_ctx = int(config.get("llama.idx.db.last"))
        last_str = "-" if last_ctx <= 0 else f"{datetime.datetime.fromtimestamp(last_ctx).strftime('%Y-%m-%d %H:%M:%S')} ({last_ctx})"
        debug.add(self.id, 'CTX [db.last]:', str(last_str))

        debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.31 23:00:00                  #
# ================================================== #

import re

def coalesce_text(parts):
    """Merge text parts, preserving intentional newlines and fixing spaces."""
    if not parts:
        return ""
    out = []
    for piece in parts:
        if not piece:
            continue
        s = str(piece)
        s = re.sub(r"[ \t\f\v]+", " ", s)
        s = re.sub(r"[ \t]*\n[ \t]*", "\n", s)
        if not out:
            out.append(s.strip())
            continue
        if out[-1].endswith("\n") or s.startswith("\n"):
            out.append(s.lstrip())
        else:
            out.append(" " + s.strip())
    text = "".join(out)
    text = re.sub(r"[ \t]+([,.;:!?%])", r"\1", text)
    text = re.sub(r"[ \t]+([\)\]\}])", r"\1", text)
    text = re.sub(r"[ \t]+(['\"])", r"\1", text)
    text = re.sub(r"\n{3,}", "\n\n", text)
    return text.strip()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.02 16:00:00                  #
# ================================================== #

from typing import Dict, Any, List

from pygpt_net.core.types import (
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    AGENT_TYPE_OPENAI,
    AGENT_TYPE_LLAMA,
)
from pygpt_net.utils import trans


class Placeholder:
    def __init__(self, window=None):
        """
        Configuration placeholder options controller

        :param window: Window instance
        """
        self.window = window
        self._apply_handlers = {
            "access_actions": lambda p: self.get_access_actions(),
            "agent_modes": lambda p: self.get_agent_modes(),
            "agent_provider": lambda p: self.get_agent_providers(),
            "agent_provider_llama": lambda p: self.get_agent_providers_llama(),
            "agent_provider_openai": lambda p: self.get_agent_providers_openai(),
            "audio_input_backend": lambda p: self.get_audio_input_backend(),
            "audio_input_devices": lambda p: self.get_audio_input_devices(),
            "audio_output_backend": lambda p: self.get_audio_output_backend(),
            "audio_output_devices": lambda p: self.get_audio_output_devices(),
            "audio_tts_whisper_voices": lambda p: self.get_audio_tts_whisper_voices(),
            "camera_devices": lambda p: self.get_camera_devices(),
            "embeddings_providers": lambda p: self.get_embeddings_providers(),
            "idx": lambda p: self.get_idx(p),
            "keys": lambda p: self.get_keys(),
            "keys_modifiers": lambda p: self.get_modifiers(),
            "langchain_providers": lambda p: self.get_langchain_providers(),
            "languages": lambda p: self.get_languages(),
            "llama_index_chat_modes": lambda p: self.get_llama_index_chat_modes(),
            "llama_index_loaders": lambda p: self.get_llama_index_loaders(),
            "llama_index_loaders_file": lambda p: self.get_llama_index_loaders(type="file"),
            "llama_index_loaders_web": lambda p: self.get_llama_index_loaders(type="web"),
            "llama_index_providers": lambda p: self.get_llama_index_providers(),
            "llm_providers": lambda p: self.get_llm_providers(),
            "models": lambda p: self.get_models(p),
            "modes": lambda p: self.get_modes(p),
            "multimodal": lambda p: self.get_multimodal(p),
            "presets": lambda p: self.get_presets(p),
            "remote_tools_openai": lambda p: self.get_remote_tools_openai(),
            "speech_synthesis_actions": lambda p: self.get_speech_synthesis_actions(),
            "styles": lambda p: self.get_styles(),
            "syntax_styles": lambda p: self.get_syntax_styles(),
            "vector_storage": lambda p: self.get_vector_storage(),
            "var_types": lambda p: self.get_var_types(),
            "voice_control_actions": lambda p: self.get_voice_control_actions(),
        }

    def _apply_combo_if_needed(self, item: Any):
        """
        Apply combo keys if needed

        :param item: item to check
        """
        if isinstance(item, dict) and item.get("type") == "combo":
            use = item.get("use")
            if use is not None:
                params = item.get("use_params")
                if not isinstance(params, dict):
                    params = {}
                item["keys"] = self.apply_by_id(use, params)

    def _apply_suboptions(self, mapping: Dict[str, Any]):
        """
        Apply placeholders to suboptions in mapping

        :param mapping: Suboptions mapping
        """
        for item in mapping.values():
            self._apply_combo_if_needed(item)

    def apply(self, option: Dict[str, Any]):
        """
        Apply placeholders to option

        :param option: Option dict
        """
        t = option["type"]
        if t == "dict" and "keys" in option:
            self._apply_suboptions(option["keys"])
        elif t == "cmd" and "params_keys" in option:
            self._apply_suboptions(option["params_keys"])
        elif t in ("combo", "bool_list"):
            use = option.get("use")
            if use is not None:
                params = option.get("use_params")
                if not isinstance(params, dict):
                    params = {}
                option["keys"] = self.apply_by_id(use, params)

    def apply_by_id(self, id: str, params: dict = None) -> List[Dict[str, str]]:
        """
        Apply placeholders by id

        :param id: Placeholder options id
        :param params: Additional parameters for specific placeholders
        :return: Filled placeholder list
        """
        if params is None:
            params = {}
        handler = self._apply_handlers.get(id)
        return handler(params) if handler else []

    def get_audio_tts_whisper_voices(self) -> List[Dict[str, str]]:
        """
        Get audio TTS whisper voices list

        :return: Filled placeholder list
        """
        voices = self.window.core.audio.whisper.get_voices()
        return [{v: v} for v in voices]

    def get_audio_input_devices(self) -> List[Dict[str, str]]:
        """
        Get audio input devices list

        :return: Filled placeholder list
        """
        devices = self.window.core.audio.get_input_devices()
        return [{str(did): name} for did, name in devices]

    def get_camera_devices(self) -> List[Dict[str, str]]:
        """
        Get camera devices list

        :return: Filled placeholder list
        """
        return self.window.core.camera.get_devices()

    def get_audio_output_devices(self) -> List[Dict[str, str]]:
        """
        Get audio output devices list

        :return: Filled placeholder list
        """
        devices = self.window.core.audio.get_output_devices()
        return [{str(did): name} for did, name in devices]

    def get_audio_input_backend(self) -> List[Dict[str, str]]:
        """
        Get audio input backends list

        :return: Filled placeholder list
        """
        items = self.window.core.audio.get_input_backends()
        return [{str(i): name} for i, name in items]

    def get_audio_output_backend(self) -> List[Dict[str, str]]:
        """
        Get audio output backends list

        :return: Filled placeholder list
        """
        items = self.window.core.audio.get_output_backends()
        return [{str(i): name} for i, name in items]

    def get_langchain_providers(self) -> List[Dict[str, str]]:
        """
        Get Langchain LLM providers list

        :return: Filled placeholder list
        """
        choices = self.window.core.llm.get_choices(MODE_LANGCHAIN)
        return [{k: v} for k, v in choices.items()]

    def get_llama_index_providers(self) -> List[Dict[str, str]]:
        """
        Get Llama-index LLM providers list

        :return: Filled placeholder list
        """
        choices = self.window.core.llm.get_choices(MODE_LLAMA_INDEX)
        return [{k: v} for k, v in choices.items()]

    def get_llm_providers(self) -> List[Dict[str, str]]:
        """
        Get all LLM provider placeholders list

        :return: Filled placeholder list
        """
        choices = self.window.core.llm.get_choices()
        return [{k: v} for k, v in choices.items()]

    def get_embeddings_providers(self) -> List[Dict[str, str]]:
        """
        Get embedding providers list

        :return: Filled placeholder list
        """
        choices = self.window.core.llm.get_choices("embeddings")
        return [{k: v} for k, v in choices.items()]

    def get_agent_providers(self) -> List[Dict[str, str]]:
        """
        Get all agent providers list

        :return: Filled placeholder list
        """
        return self.window.core.agents.provider.get_choices()

    def get_agent_providers_llama(self) -> List[Dict[str, str]]:
        """
        Get Llama-index agent provider list

        :return: Filled placeholder list
        """
        return self.window.core.agents.provider.get_choices(AGENT_TYPE_LLAMA)

    def get_agent_providers_openai(self) -> List[Dict[str, str]]:
        """
        Get OpenAI agent provider list

        :return: Filled placeholder list
        """
        return self.window.core.agents.provider.get_choices(AGENT_TYPE_OPENAI)

    def get_remote_tools_openai(self) -> List[Dict[str, str]]:
        """
        Get OpenAI remote tools list

        :return: Filled placeholder list
        """
        return self.window.core.api.openai.remote_tools.get_choices()

    def get_llama_index_chat_modes(self) -> List[Dict[str, str]]:
        """
        Get llama chat modes list

        :return: Filled placeholder list
        """
        return [
            {"best": "best"},
            {"condense_question": "condense_question"},
            {"context": "context (default)"},
            {"condense_plus_context": "condense_plus_context"},
            {"simple": "simple (no query engine)"},
            {"react": "react"},
            {"openai": "openai"},
        ]

    def get_llama_index_loaders(self, type: str = "all") -> List[Dict[str, str]]:
        """
        Get data loaders list

        :param type: data type
        :return: Filled placeholder list
        """
        data = []
        choices = self.window.controller.idx.common.get_loaders_choices()
        for choice in choices:
            if type == "all":
                data.append(choice)
            elif type == "file":
                key = next(iter(choice))
                if key.startswith("file_"):
                    data.append(choice)
            elif type == "web":
                key = next(iter(choice))
                if key.startswith("web_"):
                    data.append(choice)
        return data

    def get_vector_storage(self) -> List[Dict[str, str]]:
        """
        Get vector storage list

        :return: Filled placeholder list
        """
        ids = self.window.core.idx.storage.get_ids()
        return [{i: i} for i in ids]

    def get_var_types(self) -> List[Dict[str, str]]:
        """
        Get var types list

        :return: Filled placeholder list
        """
        types = ["str", "int", "float", "bool", "dict", "list", "None"]
        return [{t: t} for t in types]

    def get_presets(self, params: dict = None) -> List[Dict[str, str]]:
        """
        Get presets list

        :param params: Additional parameters for specific placeholders
        :return: Filled placeholder list
        """
        if params is None:
            params = {}
        presets = self.window.core.presets.get_all()
        data = [{'_': '---'}]
        data.extend([{pid: pid} for pid in presets if not str(pid).startswith("current.")])
        return data

    def get_modes(self, params: dict = None) -> List[Dict[str, str]]:
        """
        Get modes list

        :param params: Additional parameters for specific placeholders
        :return: Filled placeholder list
        """
        if params is None:
            params = {}
        modes = self.window.core.modes.get_all()
        return [{mid: trans("mode." + mid)} for mid in modes]

    def get_multimodal(self, params: dict = None) -> List[Dict[str, str]]:
        """
        Get multimodal options list

        :param params: Additional parameters for specific placeholders
        :return: Filled placeholder list
        """
        if params is None:
            params = {}
        modes = self.window.core.models.get_multimodal_list()
        return [{mid: trans("multimodal." + mid)} for mid in modes]

    def get_models(self, params: dict = None) -> List[Dict[str, str]]:
        """
        Get models list (+ provider separators)

        :param params: Additional parameters for specific placeholders
        :return: Filled placeholder list
        """
        if params is None:
            params = {}
        models = self.window.core.models.get_all()
        items: Dict[str, str] = {}
        for mid, model in models.items():
            allowed = True
            if isinstance(params.get("mode"), list):
                for m in params["mode"]:
                    if m not in model.mode:
                        allowed = False
                        break
            if allowed:
                items[mid] = model.name

        data: List[Dict[str, str]] = []
        if "allow_empty" in params and params["allow_empty"] is True:
            data.append({'_': '---'})
        providers = self.window.core.llm.get_choices()
        if not providers:
            for mid, name in sorted(items.items(), key=lambda kv: kv[1].lower()):
                data.append({mid: name})
            return data

        for provider, provider_label in providers.items():
            provider_items = [(k, v) for k, v in items.items() if models[k].provider == provider]
            if provider_items:
                data.append({f"separator::{provider}": provider_label})
                data.extend([{k: v} for k, v in provider_items])
        return data

    def get_agent_modes(self) -> List[Dict[str, str]]:
        """
        Get agent/expert modes list

        :return: Filled placeholder list
        """
        modes = self.window.core.agents.legacy.get_allowed_modes()
        return [{mid: trans(f"mode.{mid}")} for mid in modes]

    def get_languages(self) -> List[Dict[str, str]]:
        """
        Get world languages list

        :return: Filled placeholder list
        """
        return self.window.core.text.get_language_choices()

    def get_idx(self, params: dict = None) -> List[Dict[str, str]]:
        """
        Get indexes (LlamaIndex) list

        :param params: Additional parameters for specific placeholders
        :return: Filled placeholder list
        """
        if params is None:
            params = {}
        indexes = self.window.core.idx.get_idx_ids()
        data = []
        if "none" not in params or params["none"] is True:
            data.append({'_': '---'})
        for item in indexes:
            for k, v in item.items():
                data.append({k: v})
        return data

    def get_syntax_styles(self) -> List[Dict[str, str]]:
        """
        Get code syntax highlighter styles list

        :return: Filled placeholder list
        """
        styles = self.window.controller.chat.render.web_renderer.body.highlight.get_styles()
        styles.sort()
        return [{sid: sid} for sid in styles]

    def get_styles(self) -> List[Dict[str, str]]:
        """
        Get styles list (blocks, chatgpt, etc.)

        :return: Filled placeholder list
        """
        styles = self.window.controller.theme.common.get_styles_list()
        styles.sort()
        return [{sid: sid} for sid in styles]

    def get_keys(self) -> List[Dict[str, str]]:
        """
        Get keyboard keys list

        :return: Filled placeholder list
        """
        return self.window.core.access.shortcuts.get_keys_choices()

    def get_modifiers(self) -> List[Dict[str, str]]:
        """
        Get modifiers

        :return: Filled placeholder list
        """
        return self.window.core.access.shortcuts.get_modifiers_choices()

    def _translate_sort_choices(self, choices: List[Dict[str, str]]) -> List[Dict[str, str]]:
        items = [(k, trans(v)) for choice in choices for k, v in choice.items()]
        items.sort(key=lambda x: x[1])
        return [{k: v} for k, v in items]

    def get_access_actions(self) -> List[Dict[str, str]]:
        """
        Get access actions list

        :return: Filled placeholder list
        """
        choices = self.window.core.access.actions.get_access_choices()
        return self._translate_sort_choices(choices)

    def get_speech_synthesis_actions(self) -> List[Dict[str, str]]:
        """
        Get speech actions list

        :return: Filled placeholder list
        """
        choices = self.window.core.access.actions.get_speech_synthesis_choices()
        return self._translate_sort_choices(choices)

    def get_voice_control_actions(self) -> List[Dict[str, str]]:
        """
        Get voice control actions list

        :return: Filled placeholder list
        """
        choices = self.window.core.access.actions.get_voice_control_choices()
        return self._translate_sort_choices(choices)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.06 20:00:00                  #
# ================================================== #

import asyncio
from typing import Optional

from PySide6.QtCore import Slot, QRunnable, QObject, Signal

from pygpt_net.core.events import RealtimeEvent
from pygpt_net.item.ctx import CtxItem

from .options import RealtimeOptions

class RealtimeSignals(QObject):
    """Realtime signals"""
    response = Signal(object)  # RealtimeEvent

class RealtimeWorker(QRunnable):
    """
    QRunnable worker that runs a provider-specific realtime session (websocket).

    - RT_OUTPUT_READY is emitted when the audio output is ready (STREAM_BEGIN).
    - RT_OUTPUT_TEXT_DELTA is emitted for text deltas.
    - RT_OUTPUT_AUDIO_DELTA is emitted for audio chunks to be handled by the main-thread AudioDispatcher.
    - RT_OUTPUT_AUDIO_END is emitted when the session ends.
    - RT_OUTPUT_AUDIO_ERROR is emitted on error.
    """
    def __init__(
            self,
            window,
            ctx: CtxItem,
            opts: RealtimeOptions
    ):
        """
        Initialize the worker.

        :param window: Window instance
        :param ctx: CtxItem
        :param opts: RealtimeOptions
        """
        super().__init__()
        self.window = window
        self.ctx = ctx
        self.opts = opts

    def get_client(self, provider: str):
        """
        Get the appropriate client based on the provider

        :param provider: Provider name
        :return: Client instance
        """
        provider = (provider or "openai").lower()
        if provider == "google":
            return self.window.core.api.google.realtime.handler
        elif provider == "openai":
            return self.window.core.api.openai.realtime.handler
        elif provider == "x_ai":
            return self.window.core.api.xai.realtime.handler
        else:
            raise RuntimeError(f"Unsupported realtime provider: {provider}")

    @Slot()
    def run(self):
        loop = None  # ensure defined for cleanup

        # STREAM_BEGIN -> UI
        try:
            event = RealtimeEvent(RealtimeEvent.RT_OUTPUT_READY, {
                "ctx": self.ctx,
            })
            self.opts.rt_signals.response.emit(event) if self.opts.rt_signals else None
        except Exception:
            pass

        try:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            async def _amain():
                # Text deltas -> UI
                async def on_text(delta: str):
                    if not delta:
                        return
                    event = RealtimeEvent(RealtimeEvent.RT_OUTPUT_TEXT_DELTA, {
                        "ctx": self.ctx,
                        "chunk": delta,
                    })
                    self.opts.rt_signals.response.emit(event) if self.opts.rt_signals else None

                # Audio -> enqueue to main-thread
                async def on_audio(
                        data: bytes,
                        mime: str,
                        rate: Optional[int],
                        channels: Optional[int],
                        final: bool = False
                ):
                    event = RealtimeEvent(RealtimeEvent.RT_OUTPUT_AUDIO_DELTA, {
                        "payload":  {
                            "ctx": self.ctx,
                            "data": data or b"",
                            "mime": mime or "audio/pcm",
                            "rate": int(rate) if rate is not None else None,
                            "channels": int(channels) if channels is not None else None,
                            "final": bool(final),
                            "provider": self.opts.provider,
                            "model": self.opts.model,
                        }
                    })
                    self.opts.rt_signals.response.emit(event) if self.opts.rt_signals else None

                def _should_stop() -> bool:
                    try:
                        return bool(self.window.controller.kernel.stopped())
                    except Exception:
                        return False

                # run the client
                client = self.get_client(self.opts.provider)
                await client.run(self.ctx, self.opts, on_text, on_audio, _should_stop)

            loop.run_until_complete(_amain())
            # print("[rt] STREAM_END")

        except Exception as e:
            try:
                event = RealtimeEvent(RealtimeEvent.RT_OUTPUT_AUDIO_ERROR, {"error": e})
                self.opts.rt_signals.response.emit(event) if self.opts.rt_signals else None
            finally:
                pass
        finally:
            # Robust asyncio teardown to avoid hangs on subsequent runs
            if loop is not None:
                try:
                    pending = [t for t in asyncio.all_tasks(loop) if not t.done()]
                    for t in pending:
                        t.cancel()
                    if pending:
                        loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))
                except Exception:
                    pass
                try:
                    loop.run_until_complete(loop.shutdown_asyncgens())
                except Exception:
                    pass
                try:
                    loop.close()
                except Exception:
                    pass
                try:
                    asyncio.set_event_loop(None)
                except Exception:
                    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.28 20:00:00                  #
# ================================================== #

import json
from typing import List, Any, Dict, Optional

from google.genai import types as gtypes
from pygpt_net.item.model import ModelItem


class Tools:
    def __init__(self, window=None):
        """
        Tools mapper for Google GenAI

        :param window: Window instance
        """
        self.window = window

    # -------- SANITIZER --------
    def _sanitize_schema(self, schema: Any) -> Any:
        """
        Sanitize a JSON Schema dict for Google GenAI (function parameters).

        Key points:
        - Remove unsupported JSON Schema keywords (additionalProperties, oneOf, $ref, ...).
        - Normalize "type" so that it's either a single lowercase string or absent.
          Handle lists (unions), non-string types (e.g., dict), and infer a type when possible.
        - Keep "enum" only when type is string.
        - For objects, sanitize only "properties" (each property's schema) and validate "required".
        - For arrays, sanitize "items" into a single schema (object, not list).
        - Do not recurse into "properties" itself as a map, nor into "required"/"enum" as they are scalars/lists.
        """
        # 1) Fast exits
        if isinstance(schema, list):
            # Only descend into lists of dicts (complex schemas). For scalar lists (required/enum), return as is.
            if schema and all(isinstance(x, dict) for x in schema):
                return [self._sanitize_schema(x) for x in schema]
            return schema

        if not isinstance(schema, dict):
            return schema

        # 2) Remove unsupported/problematic keywords for Google function parameters
        banned = {
            "additionalProperties", "additional_properties",
            "unevaluatedProperties", "patternProperties",
            "dependencies", "dependentSchemas", "dependentRequired",
            "oneOf", "anyOf", "allOf",
            "$defs", "$ref", "$schema", "$id",
            "examples", "readOnly", "writeOnly", "nullable",
        }
        for k in list(schema.keys()):
            if k in banned:
                schema.pop(k, None)

        # 3) Normalize "type" safely
        t = schema.get("type")

        # a) If it's a list (union), pick the first non-null string, otherwise default to "object"
        if isinstance(t, list):
            t_no_null = [x for x in t if isinstance(x, str) and x.lower() != "null"]
            schema["type"] = t_no_null[0] if t_no_null else "object"
            t = schema["type"]

        # b) If "type" is not a string (could be dict or missing), try to infer; otherwise drop it
        if not isinstance(t, str):
            if isinstance(schema.get("properties"), dict):
                schema["type"] = "object"
            elif "items" in schema:
                schema["type"] = "array"
            elif isinstance(schema.get("enum"), list) and all(isinstance(x, str) for x in schema["enum"]):
                schema["type"] = "string"
            else:
                schema.pop("type", None)
        else:
            schema["type"] = t.lower()

        type_l = schema["type"].lower() if isinstance(schema.get("type"), str) else ""

        # 4) Keep enum only for string-typed schemas
        if "enum" in schema and type_l != "string":
            schema.pop("enum", None)

        # 5) Objects: sanitize properties and required
        if type_l == "object":
            props = schema.get("properties")
            if not isinstance(props, dict):
                props = {}
            clean_props: Dict[str, Any] = {}
            for pname, pval in props.items():
                clean_props[pname] = self._sanitize_schema(pval)
            schema["properties"] = clean_props

            req = schema.get("required")
            if not (isinstance(req, list) and all(isinstance(x, str) for x in req) and len(req) > 0):
                schema.pop("required", None)

        # 6) Arrays: ensure "items" is a single dict schema
        elif type_l == "array":
            items = schema.get("items")
            if isinstance(items, list):
                items = items[0] if items else {"type": "string"}
            if not isinstance(items, dict):
                items = {"type": "string"}
            schema["items"] = self._sanitize_schema(items)

        # 7) Recurse into the remaining nested dict/list values,
        #    but skip "properties", "items", "required", and "enum" (already handled)
        for k, v in list(schema.items()):
            if k in ("properties", "items", "required", "enum"):
                continue
            if isinstance(v, dict):
                schema[k] = self._sanitize_schema(v)
            elif isinstance(v, list) and v and all(isinstance(x, dict) for x in v):
                schema[k] = [self._sanitize_schema(x) for x in v]

        return schema

    # -------- CONVERTER to gtypes.Schema (UPPERCASE) --------
    def _to_gschema(self, schema: Any) -> gtypes.Schema:
        """
        Convert sanitized dict -> google.genai.types.Schema.
        Enforces UPPERCASE type names (OBJECT, ARRAY, STRING, NUMBER, INTEGER, BOOLEAN).

        :param schema: Sanitized JSON Schema as dict
        :return: gtypes.Schema
        """
        TYPE_MAP = {
            "enum": "STRING",
            "ENUM": "STRING",
            "object": "OBJECT",
            "dict": "OBJECT",
            "array": "ARRAY",
            "list": "ARRAY",
            "string": "STRING",
            "number": "NUMBER",
            "float": "NUMBER",
            "integer": "INTEGER",
            "boolean": "BOOLEAN",
            "int": "INTEGER",
            "bool": "BOOLEAN",
            "OBJECT": "OBJECT",
            "DICT": "OBJECT",
            "ARRAY": "ARRAY",
            "LIST": "ARRAY",
            "STRING": "STRING",
            "NUMBER": "NUMBER",
            "FLOAT": "NUMBER",
            "INTEGER": "INTEGER",
            "BOOLEAN": "BOOLEAN",
            "INT": "INTEGER",
            "BOOL": "BOOLEAN",
        }

        if isinstance(schema, gtypes.Schema):
            return schema

        if not isinstance(schema, dict):
            return gtypes.Schema(type="STRING")

        t = TYPE_MAP.get(str(schema.get("type", "OBJECT")).upper(), "OBJECT")
        desc = schema.get("description")
        fmt = schema.get("format")
        enum = schema.get("enum") if isinstance(schema.get("enum"), list) else None
        req = schema.get("required") if isinstance(schema.get("required"), list) else None

        gs = gtypes.Schema(
            type=t,
            description=desc,
            format=fmt,
            enum=enum,
            required=[x for x in (req or []) if isinstance(x, str)] or None,
        )

        props = schema.get("properties")
        if isinstance(props, dict):
            gs.properties = {k: self._to_gschema(v) for k, v in props.items()}

        items = schema.get("items")
        if isinstance(items, dict):
            gs.items = self._to_gschema(items)

        return gs

    def prepare(self, model: ModelItem, functions: list) -> List[gtypes.Tool]:
        """
        Prepare Google Function Declarations (types.Tool) for google-genai.

        :param model: ModelItem
        :param functions: List of function definitions as dicts with 'name', 'desc', 'params' (JSON Schema)
        :return: List of gtypes.Tool
        """
        if not functions or not isinstance(functions, list):
            return []

        fds: List[gtypes.FunctionDeclaration] = []
        for function in functions:
            name = str(function.get("name") or "").strip()
            if not name:
                continue

            desc = function.get("desc") or ""
            params: Optional[dict] = {}
            if function.get("params"):
                try:
                    params = json.loads(function["params"])
                except Exception:
                    params = {}

                params = self._sanitize_schema(params or {})
                if not params.get("type"):
                    params["type"] = "object"

            gschema = self._to_gschema(params or {"type": "object"})

            fd = gtypes.FunctionDeclaration(
                name=name,
                description=desc,
                parameters=gschema,
            )
            fds.append(fd)

        return [gtypes.Tool(function_declarations=fds)] if fds else []


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.04 00:00:00                  #
# ================================================== #

from dataclasses import dataclass
from typing import Optional, ClassVar

from .base import BaseEvent
from ...item.ctx import CtxItem


@dataclass(slots=True)
class RealtimeEvent(BaseEvent):
    """
    Realtime events

    - RT_OUTPUT_AUDIO_DELTA - audio output chunk (with payload)
    - RT_OUTPUT_READY - audio output is ready (STREAM_BEGIN)
    - RT_OUTPUT_TEXT_DELTA - text chunk (delta)
    - RT_OUTPUT_AUDIO_END - audio output ended (STREAM_END)
    - RT_OUTPUT_TURN_END - audio output turn ended (TURN_END)
    - RT_OUTPUT_AUDIO_ERROR - audio output error (STREAM_ERROR)
    - RT_OUTPUT_AUDIO_VOLUME_CHANGED - audio output volume changed (volume level)
    """
    # static id for event family
    id: ClassVar[str] = "RealtimeEvent"

    # realtime events
    RT_OUTPUT_AUDIO_DELTA = "rt.output.audio.delta"
    RT_OUTPUT_AUDIO_END = "rt.output.audio.end"
    RT_OUTPUT_AUDIO_ERROR = "rt.output.audio.error"
    RT_OUTPUT_AUDIO_VOLUME_CHANGED = "rt.output.audio.volume.changed"
    RT_OUTPUT_AUDIO_COMMIT = "rt.output.audio.commit"
    RT_OUTPUT_READY = "rt.output.audio.ready"
    RT_OUTPUT_TEXT_DELTA = "rt.output.text.delta"
    RT_OUTPUT_TURN_END = "rt.output.turn.end"
    RT_INPUT_AUDIO_DELTA = "rt.input.audio.delta"
    RT_INPUT_AUDIO_MANUAL_START = "rt.input.audio.manual.start"
    RT_INPUT_AUDIO_MANUAL_STOP = "rt.input.audio.manual.stop"